/*
 * Copyright (c) "Neo4j"
 * Neo4j Sweden AB [https://neo4j.com]
 *
 * This file is part of Neo4j.
 *
 * Neo4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package org.neo4j.cypher.internal

import org.neo4j.cypher.internal.QueryCache.CacheKey
import org.neo4j.cypher.internal.cache.CypherQueryCaches
import org.neo4j.cypher.internal.config.CypherConfiguration
import org.neo4j.cypher.internal.expressions.FunctionTypeSignature
import org.neo4j.cypher.internal.frontend.phases.BaseState
import org.neo4j.cypher.internal.options.CypherReplanOption
import org.neo4j.cypher.internal.planning.CompilationException
import org.neo4j.cypher.internal.preparser.FullyParsedQuery
import org.neo4j.cypher.internal.preparser.InputQuery
import org.neo4j.cypher.internal.preparser.PreParsedQuery
import org.neo4j.cypher.internal.runtime.InputDataStream
import org.neo4j.cypher.internal.runtime.NoInput
import org.neo4j.cypher.internal.runtime.QueryRuntimeConfig
import org.neo4j.cypher.internal.tracing.CompilationTracer
import org.neo4j.cypher.internal.tracing.CompilationTracer.QueryCompilationEvent
import org.neo4j.cypher.internal.util.InternalNotification
import org.neo4j.cypher.internal.util.InternalNotificationLogger
import org.neo4j.cypher.internal.util.RecordingNotificationLogger
import org.neo4j.dbms.systemgraph.DefaultQueryLanguageLookup
import org.neo4j.exceptions.ParameterNotFoundException
import org.neo4j.gqlstatus.ErrorGqlStatusObject
import org.neo4j.internal.kernel.api.security.AccessMode
import org.neo4j.kernel.GraphDatabaseQueryService
import org.neo4j.kernel.api.QueryLanguage
import org.neo4j.kernel.api.exceptions.Status
import org.neo4j.kernel.api.exceptions.Status.HasStatus
import org.neo4j.kernel.database.DatabaseReference
import org.neo4j.kernel.database.NamedDatabaseId
import org.neo4j.kernel.database.NormalizedCatalogEntry
import org.neo4j.kernel.database.NormalizedDatabaseName
import org.neo4j.kernel.impl.query.FunctionInformation
import org.neo4j.kernel.impl.query.FunctionInformation.InputInformation
import org.neo4j.kernel.impl.query.QueryExecution
import org.neo4j.kernel.impl.query.QueryExecutionMonitor
import org.neo4j.kernel.impl.query.QuerySubscriber
import org.neo4j.kernel.impl.query.TransactionalContext
import org.neo4j.kernel.impl.query.TransactionalContext.DatabaseMode
import org.neo4j.logging.InternalLogProvider
import org.neo4j.monitoring.Monitors
import org.neo4j.values.virtual.MapValue

import java.io.Closeable
import java.lang
import java.time.Clock
import java.util.Optional
import java.util.UUID

import scala.jdk.CollectionConverters.SeqHasAsJava
import scala.jdk.CollectionConverters.SetHasAsJava

/**
 * This class constructs and initializes both the cypher compilers and runtimes, which are very expensive
 * operation. Please make sure this will be constructed only once and properly reused.
 */
abstract class ExecutionEngine(
  val queryService: GraphDatabaseQueryService,
  val kernelMonitors: Monitors,
  val tracer: CompilationTracer,
  val config: CypherConfiguration,
  val masterCompiler: MasterCompiler,
  val queryCaches: CypherQueryCaches,
  val logProvider: InternalLogProvider,
  val clock: Clock = Clock.systemUTC()
) extends Closeable {

  // HELPER OBJECTS
  protected val defaultQueryExecutionMonitor = kernelMonitors.newMonitor(classOf[QueryExecutionMonitor])

  private val preParser = new CachingPreParser(config, queryCaches.preParserCache)

  private val defaultLanguageLookup =
    queryService.getDependencyResolver.resolveDependency(classOf[DefaultQueryLanguageLookup])

  private val queryCache: QueryCache[CacheKey[InputQuery.CacheKey], ExecutableQuery] = queryCaches.executableQueryCache

  private val schemaHelper = new SchemaHelper(queryCache, masterCompiler)

  // ACTUAL FUNCTIONALITY

  /**
   * Executes query returns a `QueryExecution` that can be used to control demand to the provided `QuerySubscriber`.
   * This method assumes this is the only query running within the transaction, and therefor will register transaction closing
   * with the TaskCloser
   *
   * @param query       the query to execute
   * @param params      the parameters of the query
   * @param context     the transactional context in which to run the query
   * @param profile     if `true` run with profiling enabled
   * @param prePopulate if `true` pre populate all results
   * @param subscriber  the subscriber where results will be streamed
   * @return a `QueryExecution` that controls the demand to the subscriber
   */
  def execute(
    query: String,
    params: MapValue,
    context: TransactionalContext,
    profile: Boolean,
    prePopulate: Boolean,
    subscriber: QuerySubscriber,
    monitor: QueryExecutionMonitor = defaultQueryExecutionMonitor
  ): QueryExecution = {
    monitor.startProcessing(context.executingQuery())
    executeSubquery(query, params, context, isOutermostQuery = true, profile, prePopulate, subscriber, monitor)
  }

  /**
   * Executes query returns a `QueryExecution` that can be used to control demand to the provided `QuerySubscriber`
   * Note. This method will monitor the query start after it has been parsed. The caller is responsible for monitoring any query failing before this point.
   *
   * @param query       the query to execute
   * @param params      the parameters of the query
   * @param context     the context in which to run the query
   * @param prePopulate if `true` pre populate all results
   * @param subscriber  the subscriber where results will be streamed
   * @return a `QueryExecution` that controls the demand to the subscriber
   */
  def execute(
    query: FullyParsedQuery,
    params: MapValue,
    context: TransactionalContext,
    prePopulate: Boolean,
    input: InputDataStream,
    queryMonitor: QueryExecutionMonitor,
    subscriber: QuerySubscriber
  ): QueryExecution = {
    queryMonitor.startProcessing(context.executingQuery())
    val queryTracer = tracer.compileQuery(query.description)
    val notificationLogger = new RecordingNotificationLogger()
    closing(context, queryTracer) {
      doExecute(
        query,
        params,
        context,
        isOutermostQuery = true,
        prePopulate,
        input,
        queryMonitor,
        queryTracer,
        subscriber,
        notificationLogger
      )
    }
  }

  /**
   * Executes query returns a `QueryExecution` that can be used to control demand to the provided `QuerySubscriber`.
   * This method assumes the query is running as one of many queries within a single transaction and therefor needs
   * to be told using the shouldCloseTransaction field if the TaskCloser needs to have a transaction close registered.
   *
   * @param query            the query to execute
   * @param params           the parameters of the query
   * @param context          the transactional context in which to run the query
   * @param isOutermostQuery provide `true` if this is the outer-most query and should close the transaction when finished or error
   * @param profile          if `true` run with profiling enabled
   * @param prePopulate      if `true` pre populate all results
   * @param subscriber       the subscriber where results will be streamed
   * @return a `QueryExecution` that controls the demand to the subscriber
   */
  def executeSubquery(
    query: String,
    params: MapValue,
    context: TransactionalContext,
    isOutermostQuery: Boolean,
    profile: Boolean,
    prePopulate: Boolean,
    subscriber: QuerySubscriber,
    monitor: QueryExecutionMonitor = defaultQueryExecutionMonitor
  ): QueryExecution = {
    val queryTracer = tracer.compileQuery(query)
    closing(context, queryTracer) {
      val couldContainSensitiveFields = isOutermostQuery && masterCompiler.supportsAdministrativeCommands()
      val notificationLogger = new RecordingNotificationLogger()
      val queryLangScope = context.kernelTransaction().defaultQueryLanguageScope()

      val defaultLanguage =
        defaultLanguageLookup.dbDefaultQueryLanguage(queryLangScope, context.databaseId(), config.systemDefaultLanguage)
      val preParsedQuery = preParser.preParseQuery(
        queryText = query,
        notificationLogger = notificationLogger,
        defaultLanguage = defaultLanguage,
        profile = profile,
        couldContainSensitiveFields = couldContainSensitiveFields,
        targetsComposite = DatabaseMode.COMPOSITE.equals(context.databaseMode())
      )
      doExecute(
        preParsedQuery,
        params,
        context,
        isOutermostQuery,
        prePopulate,
        NoInput,
        monitor,
        queryTracer,
        subscriber,
        notificationLogger
      )
    }
  }

  private def closing[T](context: TransactionalContext, traceEvent: QueryCompilationEvent)(code: => T): T =
    try code
    catch {
      case e: HasStatus =>
        context.kernelTransaction().markForTermination(e.status())
        context.close()
        throw e;

      case t: Throwable =>
        context.kernelTransaction().markForTermination(Status.Transaction.QueryExecutionFailedOnTransaction)
        context.close()
        throw t

    } finally traceEvent.close()

  private def doExecute(
    query: InputQuery,
    params: MapValue,
    context: TransactionalContext,
    isOutermostQuery: Boolean,
    prePopulate: Boolean,
    input: InputDataStream,
    queryMonitor: QueryExecutionMonitor,
    tracer: QueryCompilationEvent,
    subscriber: QuerySubscriber,
    notificationLogger: InternalNotificationLogger
  ): QueryExecution = {
    context.executingQuery().onPreparseReady(query.resolvedLanguage)
    val executableQuery =
      try {
        getOrCompile(context, query, tracer, params, notificationLogger)
      } catch {
        case gqlException: ErrorGqlStatusObject =>
          if (isOutermostQuery) {
            val status = gqlException match {
              case withStatus: HasStatus => withStatus.status()
              case _                     => null
            }
            queryMonitor.endFailure(
              context.executingQuery(),
              gqlException.getMessage,
              status,
              gqlException
            )
          }
          throw gqlException
        case up: Throwable =>
          if (isOutermostQuery) {
            val status = up match {
              case withStatus: HasStatus => withStatus.status()
              case _                     => null
            }
            queryMonitor.endFailure(context.executingQuery(), up.getMessage, status, null)
          }
          throw up
      }
    if (query.options.queryOptions.executionMode.name != "explain") {
      checkParameters(executableQuery.paramNames, params, executableQuery.extractedParams)
    }
    val combinedParams = params.updatedWith(executableQuery.extractedParams)

    if (isOutermostQuery) {
      context.executingQuery().onObfuscatorReady(executableQuery.queryObfuscator, query.options.offset.offset)
      context.executingQuery().onCompilationCompleted(
        executableQuery.compilerInfo,
        executableQuery.planDescriptionSupplier(),
        executableQuery.deprecationNotificationsProvider(query.options.offset),
        executableQuery.executionPlanCacheKeyHash
      )
    }

    val queryConfig = QueryRuntimeConfig.createFrom(query.options.queryOptions, config)
    executableQuery.execute(
      context,
      isOutermostQuery,
      query.options,
      combinedParams,
      prePopulate,
      input,
      queryMonitor,
      subscriber,
      queryConfig
    )
  }

  /*
   * Return the CompilerWithExpressionCodeGenOption to be used.
   */
  private def compilerWithExpressionCodeGenOption(
    inputQuery: InputQuery,
    tracer: QueryCompilationEvent,
    transactionalContext: TransactionalContext,
    params: MapValue,
    notificationLogger: InternalNotificationLogger,
    sessionDatabase: DatabaseReference
  ): CompilerWithExpressionCodeGenOption[ExecutableQuery] = {
    val compiledExpressionCompiler =
      () =>
        masterCompiler.compile(
          inputQuery.withRecompilationLimitReached,
          tracer,
          transactionalContext,
          params,
          notificationLogger,
          sessionDatabase
        )
    val interpretedExpressionCompiler =
      () =>
        masterCompiler.compile(
          inputQuery,
          tracer,
          transactionalContext,
          params,
          notificationLogger,
          sessionDatabase
        )

    new CompilerWithExpressionCodeGenOption[ExecutableQuery] {
      override def compile(): ExecutableQuery = {
        if (inputQuery.options.compileWhenHot && config.recompilationLimit == 0) {
          // We have recompilationLimit == 0, go to compiled directly
          compiledExpressionCompiler()
        } else {
          interpretedExpressionCompiler()
        }
      }

      override def compileWithExpressionCodeGen(): ExecutableQuery = compiledExpressionCompiler()

      override def maybeCompileWithExpressionCodeGen(
        hitCount: Int,
        shouldRecompile: () => Boolean
      ): Option[ExecutableQuery] = {
        // check if we need to do jit compiling of queries and if hot enough
        if (
          inputQuery.options.compileWhenHot && config.recompilationLimit > 0 && hitCount >= config.recompilationLimit && shouldRecompile()
        ) {
          Some(compiledExpressionCompiler())
        } else {
          // In the other case we have no recompilation step
          None
        }
      }
    }
  }

  private def getOrCompile(
    context: TransactionalContext,
    initialInputQuery: InputQuery,
    tracer: QueryCompilationEvent,
    params: MapValue,
    notificationLogger: InternalNotificationLogger
  ): ExecutableQuery = {

    // create transaction and query context
    val tc = context.getOrBeginNewIfClosed()
    val compilerAuthorization = tc.restrictCurrentTransaction(tc.securityContext.withMode(AccessMode.Static.READ))
    var forceReplan = false
    var inputQuery = initialInputQuery

    val cacheKey = CacheKey(
      inputQuery.cacheKey,
      QueryCache.extractParameterTypeMap(params, config.useParameterSizeHint),
      tc.kernelTransaction().dataRead().transactionStateHasChanges(),
      inputQuery.resolvedLanguage
    )

    try {
      var n = 0
      while (n < ExecutionEngine.PLAN_BUILDING_TRIES) {

        val schemaToken = schemaHelper.readSchemaToken(tc)
        if (forceReplan) {
          forceReplan = false
          inputQuery = inputQuery.withReplanOption(CypherReplanOption.force)
        }
        val compiler = compilerWithExpressionCodeGenOption(
          inputQuery,
          tracer,
          tc,
          params,
          notificationLogger,
          new DatabaseReference {
            override def alias(): NormalizedDatabaseName = throw new NotImplementedError()

            /**
           * @return the namespace that the alias is in, or empty if it is in the default namespace
           */
            override def namespace(): Optional[NormalizedDatabaseName] = throw new NotImplementedError()

            /**
           * @return whether the alias associated with this reference is the database's original/true name
           */
            override def isPrimary: Boolean = throw new NotImplementedError()

            /**
           * @return the unique identity for this reference
           */
            override def id(): UUID = throw new NotImplementedError()

            override def namedDatabaseId(): NamedDatabaseId = throw new NotImplementedError()

            /**
           * @return Prettified String representaion
           */
            override def toPrettyString: String = throw new NotImplementedError()

            /**
           * @return the full normalized name of the dataspace, including the namespace.
           */
            override def fullName(): NormalizedDatabaseName = context.databaseId().normalizedName()

            /**
           * @return true if this reference points to a Composite database, otherwise false
           */
            override def isComposite: Boolean = context.databaseMode().equals(DatabaseMode.COMPOSITE)

            override def compareTo(o: DatabaseReference): Int = throw new NotImplementedError()

            /**
             * @return the owning database of this reference. This is used for authorization on property shards which inherit
             *         their permissions from the "owning" graph shard.
             */
            override def owningDatabaseName: String = fullName().name();

            override def catalogEntry(): NormalizedCatalogEntry = throw new NotImplementedError()

            override def isShard: Boolean = false
          }
        )

        val executableQuery = queryCache.computeIfAbsentOrStale(
          cacheKey,
          tc,
          compiler,
          inputQuery.options.queryOptions.replan,
          context.executingQuery().id()
        )

        val lockedEntities = schemaHelper.lockEntities(schemaToken, executableQuery, tc)

        if (lockedEntities.successful) {
          return executableQuery
        }
        forceReplan = lockedEntities.needsReplan

        // if the schema has changed while taking all locks we need to try again.
        n += 1
      }
    } finally {
      compilerAuthorization.close()
    }

    throw CompilationException.tooFrequentSchemaChanges()
  }

  def clearQueryCaches(): Long =
    List(masterCompiler.clearCaches(), queryCache.clear(), preParser.clearCache()).max

  def clearPreParserCache(): Long =
    preParser.clearCache()

  def clearExecutableQueryCache(): Long =
    queryCache.clear()

  def clearCompilerCaches(): Long =
    masterCompiler.clearCaches()

  def insertIntoCache(
    preParsedQuery: PreParsedQuery,
    params: MapValue,
    parsedQuery: BaseState,
    parsingNotifications: Set[InternalNotification]
  ): Unit = {
    preParser.insertIntoCache(preParsedQuery)
    masterCompiler.insertIntoCache(preParsedQuery, params, parsedQuery, parsingNotifications)
  }

  def getCypherFunctions: java.util.List[FunctionInformation] = {
    val informations: Seq[FunctionInformation] =
      org.neo4j.cypher.internal.expressions.functions.Function.functionInfo.map(FunctionWithInformation)
    val predicateInformations: Seq[FunctionInformation] =
      org.neo4j.cypher.internal.expressions.IterablePredicateExpression.functionInfo.map(FunctionWithInformation)
    (informations ++ predicateInformations).asJava
  }

  override def close(): Unit =
    queryCaches.close()

  // HELPERS

  @throws(classOf[ParameterNotFoundException])
  private def checkParameters(queryParams: Array[String], givenParams: MapValue, extractedParams: MapValue): Unit = {
    var i = 0
    while (i < queryParams.length) {
      val key = queryParams(i)
      if (!(givenParams.containsKey(key) || extractedParams.containsKey(key))) {
        val missingKeys =
          queryParams.filter(key => !(givenParams.containsKey(key) || extractedParams.containsKey(key))).distinct
        throw ParameterNotFoundException.expectedParam(
          missingKeys.mkString(", "),
          givenParams.keySet()
        )
      }
      i += 1
    }
  }
}

case class FunctionWithInformation(f: FunctionTypeSignature) extends FunctionInformation {

  override def getFunctionName: String = f.function.name

  override def getDescription: String = f.description

  override def getCategory: String = f.category

  override def getSignature: String = f.getSignatureAsString

  override def isAggregationFunction: lang.Boolean = f.isAggregationFunction

  override def isDeprecated: lang.Boolean = f.deprecated

  override def deprecatedBy: java.util.Optional[lang.String] = Optional.ofNullable(f.deprecatedBy.orNull)

  override def returnType: String = f.outputType.normalizedCypherTypeString()

  override def inputSignature: java.util.List[InputInformation] = {
    f.names.zip(f.argumentTypes ++ f.optionalTypes).map { case (name, cType) =>
      val typeString = f.overriddenArgumentTypeName match {
        case Some(map) => map.getOrElse(name, cType.normalizedCypherTypeString())
        case None      => cType.normalizedCypherTypeString()
      }
      val argumentDescription = f.argumentDescriptions.getOrElse(name, s"$name :: $typeString")
      new InputInformation(
        name,
        typeString,
        argumentDescription,
        false,
        java.util.Optional.empty[String]()
      )
    }.asJava
  }

  override def scopes(): java.util.Set[QueryLanguage] = {
    f.scopes.map(org.neo4j.cypher.internal.frontend.phases.QueryLanguage.toKernelScope).asJava
  }
}

object ExecutionEngine {
  val PLAN_BUILDING_TRIES: Int = 20
}
