/*
 * Copyright (c) "Neo4j"
 * Neo4j Sweden AB [https://neo4j.com]
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.neo4j.cypher.internal.ast

import org.neo4j.cypher.internal.CypherVersion
import org.neo4j.cypher.internal.ast.ConditionalQueryWhen.msg
import org.neo4j.cypher.internal.ast.ConditionalQueryWhen.name
import org.neo4j.cypher.internal.ast.ReturnItems.ReturnVariables
import org.neo4j.cypher.internal.ast.Union.UnionMapping
import org.neo4j.cypher.internal.ast.UnmappedUnion.errorParam
import org.neo4j.cypher.internal.ast.prettifier.ExpressionStringifier
import org.neo4j.cypher.internal.ast.semantics.Scope
import org.neo4j.cypher.internal.ast.semantics.SemanticAnalysisTooling
import org.neo4j.cypher.internal.ast.semantics.SemanticAnalysisToolingErrorWithGqlInfo
import org.neo4j.cypher.internal.ast.semantics.SemanticCheck
import org.neo4j.cypher.internal.ast.semantics.SemanticCheck.fromFunctionWithContext
import org.neo4j.cypher.internal.ast.semantics.SemanticCheck.fromState
import org.neo4j.cypher.internal.ast.semantics.SemanticCheck.success
import org.neo4j.cypher.internal.ast.semantics.SemanticCheck.when
import org.neo4j.cypher.internal.ast.semantics.SemanticCheckContext
import org.neo4j.cypher.internal.ast.semantics.SemanticCheckResult
import org.neo4j.cypher.internal.ast.semantics.SemanticCheckable
import org.neo4j.cypher.internal.ast.semantics.SemanticError
import org.neo4j.cypher.internal.ast.semantics.SemanticExpressionCheck
import org.neo4j.cypher.internal.ast.semantics.SemanticFeature
import org.neo4j.cypher.internal.ast.semantics.SemanticState
import org.neo4j.cypher.internal.ast.semantics.Symbol
import org.neo4j.cypher.internal.expressions.Expression
import org.neo4j.cypher.internal.expressions.ExpressionWithComputedDependencies
import org.neo4j.cypher.internal.expressions.LogicalVariable
import org.neo4j.cypher.internal.expressions.Variable
import org.neo4j.cypher.internal.util.ASTNode
import org.neo4j.cypher.internal.util.InputPosition
import org.neo4j.cypher.internal.util.Rewriter
import org.neo4j.cypher.internal.util.SubqueryVariableShadowing
import org.neo4j.cypher.internal.util.symbols.CTBoolean
import org.neo4j.cypher.internal.util.topDown
import org.neo4j.gqlstatus.GqlHelper

sealed trait QueryUtils {

  /**
   * All variables that are explicitly listed to be returned from this statement.
   * This also includes the information whether all other potentially existing variables in scope are also returned.
   */
  def returnVariables: ReturnVariables

  /**
   * All variables that are explicitly listed to be returned from this statement.
   */
  def returnColumns: List[LogicalVariable] = returnVariables.explicitVariables.toList

  /**
   * True iff this query part ends with a return clause.
   */
  def isReturning: Boolean

  /**
   * Given the root scope for this query part,
   * looks up the final scope after the last clause
   */
  def finalScope(scope: Scope): Scope

  /**
   * Semantic check for when this `Query` is in a subquery, and might import
   * variables from the `outer` scope
   */
  def semanticCheckInSubqueryContext(outer: SemanticState, current: SemanticState): SemanticCheck
  def semanticCheckImportingWithSubQueryContext(outer: SemanticState): SemanticCheck

  /**
   * Exists and Count can omit the Return Statement
   * Count still requires it for Distinct Unions as in this case the count
   * changes based on which rows are distinct vs not
   */
  def semanticCheckInSubqueryExpressionContext(
    canOmitReturn: Boolean,
    outer: SemanticState,
    context: UnaliasedNotAllowed = ImportingWithSubqueryCall
  ): SemanticCheck

  /**
   * Semantic check for when this `Query` is enclosed in Top Level Braces
   */
  def semanticCheckInTopLevelBracesContext: SemanticCheck

  def semanticCheckInConditionalQueryContext: SemanticCheck

}

sealed trait Query extends Statement with SemanticCheckable with SemanticAnalysisTooling with QueryUtils {
  def containsUpdates: Boolean

  /**
   * All Return clauses contained within this statement.
   */
  def getReturns: Seq[Return]

  /**
   * True iff this query part ends with a finish clause.
   */
  def endsWithFinish: Boolean

  /**
   * Check this query part if it starts with an importing WITH
   */
  def checkImportingWith: SemanticCheck

  /**
   * True if this query part starts with an importing WITH (has incoming arguments)
   */
  def isCorrelated: Boolean

  /**
   * Returns names of variables imported using importing WITH
   */
  def importColumns: Seq[String]

  /**
   * Return a copy of this query where the mapping function f is applied
   * to each single query, regardless if this a single query or a union query.
   */
  def mapEachSingleQuery(f: SingleQuery => SingleQuery): Query

  /**
   * Used in UnwrapTopLevelBraces
   */
  def getQuery(fromUnion: Boolean): Query

  /**
   * Return a list of all CommandClause's existing in this query.
   */
  def getCommandClauses: Seq[CommandClause]
}

sealed trait PartQuery extends Query {
  def clauses: Seq[Clause]

  def singleQuery: SingleQuery

  override def containsUpdates: Boolean =
    clauses.exists {
      case sub: SubqueryCall => sub.innerQuery.containsUpdates
      case call: CallClause  => !call.containsNoUpdates
      case _: UpdateClause   => true
      case _                 => false
    }

  override def returnVariables: ReturnVariables = clauses.last.returnVariables

  override def getReturns: Seq[Return] = {
    clauses.last match {
      case r: Return => Seq(r)
      case _         => Seq.empty
    }
  }

  override def isReturning: Boolean = clauses.last match {
    case _: Return => true
    case _         => false
  }

  override def endsWithFinish: Boolean = clauses.last match {
    case _: Finish => true
    case _         => false
  }

  override def finalScope(scope: Scope): Scope =
    if (scope.children.size < 1) Scope.empty else scope.children.last

}

case class SingleQuery(clauses: Seq[Clause])(val position: InputPosition) extends PartQuery {
  assert(clauses.nonEmpty)

  override def singleQuery: SingleQuery = this

  val getCommandClauses: Seq[CommandClause] =
    clauses.filter(_.isInstanceOf[CommandClause]).map(_.asInstanceOf[CommandClause])

  override def mapEachSingleQuery(f: SingleQuery => SingleQuery): Query = f(this)

  override def getQuery(fromUnion: Boolean): Query = this

  lazy val partitionedClauses: SingleQuery.PartitionedClauses = SingleQuery.partitionClauses(clauses)

  /**
   * The query is correlated if it imports variables from a parent query, this can happen if:
   *   - it contains an importing `WITH` clause (in first position or in second position right after a `USE` clause), or
   *   - it starts with a dynamic `USE` clause where the graph reference depends on a variable when invoking `graph.byName` or `graph.byElementId` (in a composite query).
   */
  override def isCorrelated: Boolean =
    partitionedClauses.importingWith.isDefined ||
      partitionedClauses.initialGraphSelection.exists(_.graphReference.dependencies.nonEmpty)

  override def importColumns: Seq[String] = partitionedClauses.importingWith match {
    case Some(w) => w.returnItems.items.map(_.name)
    case _       => Seq.empty
  }

  private def leadingNonImportingWith: Option[With] =
    if (partitionedClauses.importingWith.isDefined)
      None
    else
      partitionedClauses.clausesExceptImportingWithAndLeadingGraphSelection.headOption match {
        case Some(nonImportingWith: With) => Some(nonImportingWith)
        case _                            => None
      }

  private def semanticCheckAbstractInScopeSubquery(
    clauses: Seq[Clause],
    clauseCheck: Seq[Clause] => SemanticCheck,
    canOmitReturnClause: Boolean = false
  ): SemanticCheck =
    checkStandaloneCall(clauses) chain
      withScopedState(clauseCheck(clauses)) chain
      checkComposableNonTransactionCommandsAllowed() chain
      checkOrder(clauses, canOmitReturnClause) chain
      checkNoCallInTransactionsAfterWriteClause(clauses) chain
      checkInputDataStream(clauses) chain
      checkUsePositionInScopeSubquery() chain
      recordCurrentScope(this)

  private def semanticCheckAbstract(
    clauses: Seq[Clause],
    clauseCheck: Seq[Clause] => SemanticCheck,
    canOmitReturnClause: Boolean = false
  ): SemanticCheck =
    checkStandaloneCall(clauses) chain
      withScopedState(clauseCheck(clauses)) chain
      checkComposableNonTransactionCommandsAllowed() chain
      checkOrder(clauses, canOmitReturnClause) chain
      checkNoCallInTransactionsAfterWriteClause(clauses) chain
      checkInputDataStream(clauses) chain
      checkUsePosition() chain
      recordCurrentScope(this)

  override def semanticCheck: SemanticCheck =
    semanticCheckAbstract(clauses, checkClauses(_, None))

  override def semanticCheckInTopLevelBracesContext: SemanticCheck =
    semanticCheckAbstract(clauses, checkClauses(_, None, TopLevelBraces))

  override def semanticCheckInConditionalQueryContext: SemanticCheck =
    semanticCheckAbstract(clauses, checkClauses(_, None, ConditionalQueryWhen))

  /**
   * No outer scope is needed for checkClauses as we don't need to check the naming of any returned variables
   * as no variables from EXISTS / COUNT can be returned, unlike in CALL subqueries.
   */
  override def semanticCheckInSubqueryExpressionContext(
    canOmitReturn: Boolean,
    outer: SemanticState,
    context: UnaliasedNotAllowed
  ): SemanticCheck =
    semanticCheckAbstract(clauses, checkClauses(_, None, context), canOmitReturnClause = canOmitReturn)

  override def checkImportingWith: SemanticCheck = partitionedClauses.importingWith.foldSemanticCheck(_.semanticCheck)

  override def semanticCheckImportingWithSubQueryContext(outer: SemanticState): SemanticCheck = {
    def importVariables: SemanticCheck =
      partitionedClauses.importingWith.foldSemanticCheck(wth =>
        wth.semanticCheckContinuation(outer.currentScope.scope) chain
          recordCurrentScope(wth)
      )

    val workingGraph = outer.workingGraph

    checkIllegalImportWith chain
      checkInitialGraphSelection(outer) chain
      semanticCheckAbstract(
        partitionedClauses.clausesExceptImportingWithAndInitialGraphSelection,
        importVariables chain checkClauses(_, Some(outer.currentScope.scope), ImportingWithSubqueryCall)
      ) chain
      warnOnPotentiallyShadowVariables(outer) chain
      SemanticCheck.fromState(state =>
        SemanticCheck.setState(state.recordWorkingGraph(workingGraph))
      ) // resetWorkingGraph
  }

  override def semanticCheckInSubqueryContext(outer: SemanticState, current: SemanticState): SemanticCheck = {
    val workingGraph = outer.workingGraph

    checkInitialGraphSelection(outer) chain
      semanticCheckAbstractInScopeSubquery(
        partitionedClauses.clausesExceptInitialGraphSelection,
        checkClauses(_, Some(outer.currentScope.scope), ScopeClauseSubqueryCall)
      ) chain
      errorOnShadowedImportVariables(outer) chain
      warnOnPotentiallyShadowVariables(current) chain
      SemanticCheck.fromState(state =>
        SemanticCheck.setState(state.recordWorkingGraph(workingGraph))
      ) // resetWorkingGraph
  }

  private def checkInitialGraphSelection(outer: SemanticState): SemanticCheck =
    partitionedClauses.initialGraphSelection.foldSemanticCheck { graphSelection =>
      withState(outer)(graphSelection.semanticCheck)
    }

  private def checkIllegalImportWith: SemanticCheck = leadingNonImportingWith.foldSemanticCheck { wth =>
    def err(keyword: String): SemanticCheck =
      error(SemanticError.invalidImportingWithKeyword(keyword, wth.position))

    def checkReturnItems: SemanticCheck = {
      val invalidValues = wth.returnItems.items.find(!_.isPassThrough)
      when(invalidValues.nonEmpty) {
        val value = invalidValues.head
        val aliasString = if (value.alias.nonEmpty) s" AS ${value.alias.get.name}" else ""
        val expression = ExpressionStringifier.apply().apply(value.expression)
        val input = expression + aliasString
        error(SemanticError.invalidImportingWithAlisOrExpression(input, wth.position))
      }
    }

    def checkDistinct: SemanticCheck = when(wth.distinct) { err("DISTINCT") }
    def checkOrderBy: SemanticCheck = wth.orderBy.foldSemanticCheck(_ => err("ORDER BY"))
    def checkWhere: SemanticCheck = wth.where.foldSemanticCheck(_ => err("WHERE"))
    def checkSkip: SemanticCheck = wth.skip.foldSemanticCheck(_ => err("SKIP"))
    def checkLimit: SemanticCheck = wth.limit.foldSemanticCheck(_ => err("LIMIT"))

    fromState { state =>
      val resultState = wth.returnItems.items.foldSemanticCheck(_.semanticCheck)
        .run(state, SemanticCheckContext.empty)

      // [[ExpressionWithComputedDependencies]] do not carry their dependencies directly. Instead the dependencies are stored in the recorded scopes in the semantic state.
      // See also: [[computeDependenciesForExpressions]]
      val rewriter = topDown(Rewriter.lift {
        case x: ExpressionWithComputedDependencies =>
          val dependencies =
            resultState.state.recordedScopes(x.subqueryAstNode).declarationsAndDependencies.dependencies
          x.withComputedScopeDependencies(dependencies.map(_.asVariable))
      })

      val hasImports = wth.returnItems.includeExisting || wth.returnItems.items.exists { item =>
        rewriter.apply(item.expression).asInstanceOf[Expression].dependencies.nonEmpty
      }
      when(hasImports) {
        checkReturnItems chain
          checkDistinct chain
          checkWhere chain
          checkOrderBy chain
          checkSkip chain
          checkLimit
      }
    }
  }

  private def checkStandaloneCall(clauses: Seq[Clause]): SemanticCheck = {
    clauses match {
      case Seq(_: UnresolvedCall, where: With) =>
        val gql = GqlHelper.getGql42001_42N24(
          "CALL",
          "WHERE",
          where.position.offset,
          where.position.line,
          where.position.column
        )
        error(
          gql,
          "Cannot use standalone call with WHERE (instead use: `CALL ... WITH * WHERE ... RETURN *`)",
          where.position
        )
      case Seq(_: GraphSelection, _: UnresolvedCall) =>
        // USE clause and standalone procedure call
        success
      case all if all.size > 1 && all.exists(c => c.isInstanceOf[UnresolvedCall]) =>
        // Non-standalone procedure call should not allow YIELD *
        clauses.find {
          case uc: UnresolvedCall => uc.yieldAll
          case _                  => false
        }.map(c =>
          error(SemanticError.invalidYieldStar(c.position))
        )
          .getOrElse(success)
      case _ =>
        success
    }
  }

  private def checkComposableNonTransactionCommandsAllowed(): SemanticCheck = {
    // Combining commands other than show and terminate transactions are hidden behind a feature flag
    if (getCommandClauses.size > 1) {
      val nonTransactionCommands = getCommandClauses.filter(c => !c.isInstanceOf[TransactionsCommandClause])

      if (nonTransactionCommands.nonEmpty) {
        requireFeatureSupport(
          "Composing commands other than `SHOW TRANSACTIONS` and `TERMINATE TRANSACTIONS`",
          SemanticFeature.ComposableCommands,
          position
        )
      } else {
        success
      }
    } else {
      success
    }
  }

  private def checkOrder(clauses: Seq[Clause], canOmitReturnClause: Boolean): SemanticCheck = {
    fromFunctionWithContext { (s: SemanticState, c: SemanticCheckContext) =>
      {
        val sequenceErrors = clauses.sliding(2).foldLeft(Vector.empty[SemanticError]) {
          case (semanticErrors, pair) =>
            val optError = pair match {
              case Seq(match1: Match, match2: Match) if match1.optional && !match2.optional =>
                Some(SemanticError.invalidUseOfMatch(match2.position))
              case Seq(clause: Return, _) =>
                Some(SemanticError.invalidUseOfReturn(clause.name, clause.position))
              case Seq(clause: Finish, _) =>
                Some(SemanticError(s"${clause.name} can only be used at the end of the query.", clause.position))
              case Seq(_: UpdateClause, _: UpdateClause) =>
                None
              case Seq(_: UpdateClause, _: With) =>
                None
              case Seq(_: UpdateClause, _: Return) =>
                None
              case Seq(_: UpdateClause, _: Finish) =>
                None
              case Seq(update: UpdateClause, clause) if c.cypherVersion == CypherVersion.Cypher5 =>
                Some(SemanticError.withIsRequiredBetween(update.name, clause.name, clause.position))
              case _ =>
                None
            }
            optError.fold(semanticErrors)(semanticErrors :+ _)
        }

        val commandErrors =
          if (getCommandClauses.size > 1) {
            val missingYield = clauses.sliding(2).foldLeft(Vector.empty[SemanticError]) {
              case (semanticErrors, pair) =>
                val optError = pair match {
                  case Seq(command: CommandClause, clause: With) if command.yieldAll =>
                    Some(SemanticError.invalidYieldStar(
                      command.name,
                      clause.position
                    ))
                  case Seq(_: CommandClause, clause: With) if clause.withType == ParsedAsYield => None
                  case Seq(command: CommandClause, _) =>
                    Some(SemanticError.missingYield(
                      command.name,
                      command.position
                    ))
                  case _ => None
                }
                optError.fold(semanticErrors)(semanticErrors :+ _)
            }

            val missingReturn = clauses.last match {
              case clause: Return if !clause.addedInRewrite => None
              case clause =>
                Some(SemanticError.missingReturn(
                  clause.position
                ))
            }

            missingYield ++ missingReturn
          } else Vector.empty[SemanticError]

        val concludeError = clauses match {
          // standalone procedure call
          case Seq(_: CallClause)                    => None
          case Seq(_: GraphSelection, _: CallClause) => None

          case Seq() => Some(SemanticError.queryMustConcludeWithClause(this.position))

          // otherwise
          case seq => seq.last match {
              case _: UpdateClause | _: Return | _: Finish | _: CommandClause                                  => None
              case subquery: SubqueryCall if !subquery.innerQuery.isReturning && subquery.reportParams.isEmpty => None
              case call: CallClause if call.returnVariables.explicitVariables.isEmpty && !call.yieldAll        => None
              case call: CallClause         => Some(SemanticError.queryCannotConcludeWithCall(call.name, call.position))
              case _ if canOmitReturnClause => None
              case clause => Some(SemanticError.queryCannotConcludeWithClause(clause.name, clause.position))
            }
        }

        SemanticCheckResult(s, sequenceErrors ++ concludeError ++ commandErrors)
      }
    }
  }

  private def checkNoCallInTransactionsAfterWriteClause(clauses: Seq[Clause]): SemanticCheck = {
    case class Acc(precedingWrite: Boolean, errors: Seq[SemanticError])

    val Acc(_, errors) = clauses.foldLeft[Acc](Acc(precedingWrite = false, Seq.empty)) {
      case (Acc(precedingWrite, errors), callInTxs: SubqueryCall) if SubqueryCall.isTransactionalSubquery(callInTxs) =>
        if (precedingWrite) {
          Acc(
            precedingWrite,
            errors :+ SemanticError.invalidUseOfCIT(callInTxs.position)
          )
        } else {
          Acc(precedingWrite, errors)
        }
      case (acc, clause) => Acc(
          acc.precedingWrite || clause.folder.treeExists { case _: UpdateClause => true },
          acc.errors
        )
    }
    errors
  }

  private def checkClauses(
    clauses: Seq[Clause],
    outerScope: Option[Scope],
    context: UnaliasedNotAllowed = ImportingWithSubqueryCall
  ): SemanticCheck = {
    val lastIndex = clauses.size - 1
    clauses.zipWithIndex.foldSemanticCheck {
      case (clause, idx) =>
        val next = SemanticCheck.fromState { state =>
          clause match {
            case c: Return =>
              checkReturn(c, outerScope, context)
            case c: HorizonClause =>
              checkHorizon(c, outerScope)
            case _ =>
              clause.semanticCheck.map { checked =>
                val resultState = clause match {
                  case _: UpdateClause if idx == lastIndex =>
                    checked.state.newSiblingScope
                  case cc: CallClause
                    if cc.returnVariables.explicitVariables.isEmpty && !cc.yieldAll && idx == lastIndex =>
                    checked.state.newSiblingScope
                  case _ =>
                    checked.state
                }
                checked.copy(state = resultState)
              }
          }
        }

        next chain recordCurrentScope(clause)
    }
  }

  private def checkReturn(
    clause: Return,
    outerScope: Option[Scope],
    context: UnaliasedNotAllowed
  ): SemanticCheck = {
    val returnWithContext = clause.copy(context = context)(clause.position)
    for {
      closingResult <- returnWithContext.semanticCheck
      continuationResult <-
        returnWithContext.semanticCheckContinuation(closingResult.state.currentScope.scope, outerScope)
    } yield {
      semantics.SemanticCheckResult(continuationResult.state, closingResult.errors ++ continuationResult.errors)
    }

  }

  private def checkHorizon(clause: HorizonClause, outerScope: Option[Scope]): SemanticCheck = {
    for {
      closingResult <- clause.semanticCheck
      continuationResult <- clause.semanticCheckContinuation(closingResult.state.currentScope.scope, outerScope)
    } yield {
      semantics.SemanticCheckResult(continuationResult.state, closingResult.errors ++ continuationResult.errors)
    }
  }

  private def checkInputDataStream(clauses: Seq[Clause]): SemanticCheck = {
    val idsClauses = clauses.filter(_.isInstanceOf[InputDataStream])

    idsClauses.size match {
      case c if c > 1 =>
        error("There can be only one INPUT DATA STREAM in a query", idsClauses(1).position)
      case c if c == 1 =>
        if (clauses.head.isInstanceOf[InputDataStream]) {
          success
        } else {
          error("INPUT DATA STREAM must be the first clause in a query", idsClauses.head.position)
        }
      case _ => success
    }
  }

  private def checkUsePositionInScopeSubquery(): SemanticCheck = {
    val clauses = partitionedClauses.clausesExceptInitialGraphSelection

    clauses.collect {
      case useGraph: UseGraph => useGraph
    }.foldSemanticCheck { clause =>
      SemanticAnalysisToolingErrorWithGqlInfo.invalidPlacementOfUseClauseError(
        clause.position
      )
    }
  }

  private def checkUsePosition(): SemanticCheck = {
    val clauses = partitionedClauses.clausesExceptImportingWithAndLeadingGraphSelection

    clauses.collect {
      case useGraph: UseGraph => useGraph
    }.foldSemanticCheck { clause =>
      error(SemanticError.invalidPlacementOfUseClauseVerboseLegacyMsg(clause.position))
    }
  }

  private def warnOnPotentiallyShadowVariables(outer: SemanticState): SemanticCheck = {
    (inner: SemanticState) =>
      val outerScopeSymbols: Map[String, Symbol] = outer.currentScope.scope.symbolTable
      val innerScopeSymbols: Map[String, Set[Symbol]] = inner.currentScope.scope.allSymbols

      def isShadowed(s: Symbol): Boolean =
        innerScopeSymbols.contains(s.name) &&
          !innerScopeSymbols(s.name).map(_.definition).contains(s.definition)

      val shadowedSymbols = outerScopeSymbols.collect {
        case (symbolName, symbol) if isShadowed(symbol) =>
          symbolName -> innerScopeSymbols(
            symbolName
          ).find(_.definition != symbol.definition).get.definition.asVariable.position
      }
      val stateWithNotifications = shadowedSymbols.foldLeft(inner) {
        case (state, (varName, pos)) =>
          state.addNotification(SubqueryVariableShadowing(pos, varName))
      }

      SemanticCheckResult.success(stateWithNotifications)
  }

  private def errorOnShadowedImportVariables(outer: SemanticState): SemanticCheck = { (inner: SemanticState) =>
    val outerScopeSymbols: Map[String, Symbol] = outer.currentScope.scope.symbolTable

    // Finds symbols of children of innerScope
    val childrenTables = inner.currentScope.scope.children.map(_.symbolTable)
    val innerScopeSymbols: Map[String, Set[Symbol]] =
      childrenTables.foldLeft(Map.empty[String, Set[Symbol]]) {
        case (acc0, table) =>
          table.foldLeft(acc0) {
            case (acc, (str, symbol)) if acc.contains(str) =>
              acc.updated(str, acc(str) + symbol)
            case (acc, (str, symbol)) =>
              acc.updated(str, Set(symbol))
          }
      }

    def isShadowed(s: Symbol): Boolean = {
      innerScopeSymbols.contains(s.name) &&
      !innerScopeSymbols(s.name).map(_.definition).forall(_ == s.definition)
    }

    val shadowedSymbols = outerScopeSymbols.collect {
      case (symbolName, symbol) if isShadowed(symbol) =>
        symbolName -> innerScopeSymbols(
          symbolName
        ).find(_.definition != symbol.definition).get.definition.asVariable.position
    }

    val shadowingErrors = shadowedSymbols.map {
      case (varName, pos) =>
        SemanticError(
          s"The variable `$varName` is shadowing an imported variable with the same name and needs to be renamed",
          pos
        )
    }.toSeq

    SemanticCheckResult(inner, shadowingErrors)
  }
}

object SingleQuery {

  /**
   * The clauses making up a single query.
   *
   * @param initialGraphSelection A `USE` clause in first position.
   * @param importingWith An importing `WITH` clause in first position, or in second position immediately after [[initialGraphSelection]] when defined.
   * @param subsequentGraphSelection A `USE` clause in second position immediately after [[importingWith]]. If this field is defined then [[initialGraphSelection]] cannot be.
   * @param clausesExceptImportingWithAndLeadingGraphSelection All the other clauses afterwards.
   */
  case class PartitionedClauses(
    initialGraphSelection: Option[GraphSelection],
    importingWith: Option[With],
    subsequentGraphSelection: Option[GraphSelection],
    clausesExceptImportingWithAndLeadingGraphSelection: Seq[Clause]
  ) {

    lazy val leadingGraphSelection: Option[GraphSelection] =
      initialGraphSelection.orElse(subsequentGraphSelection)

    lazy val clausesExceptImportingWithAndInitialGraphSelection: Seq[Clause] =
      subsequentGraphSelection.toSeq ++ clausesExceptImportingWithAndLeadingGraphSelection

    lazy val clausesExceptInitialGraphSelection: Seq[Clause] =
      importingWith.toSeq ++ subsequentGraphSelection ++ clausesExceptImportingWithAndLeadingGraphSelection
  }

  private def partitionClauses(clauses: Seq[Clause]): PartitionedClauses =
    startingWithImportingWith(clauses)
      .orElse(startingWithGraphSelection(clauses))
      .getOrElse(PartitionedClauses(None, None, None, clauses))

  private def startingWithImportingWith(clauses: Seq[Clause]): Option[PartitionedClauses] =
    extractImportingWith(clauses).map { case (importingWith, subsequentClauses) =>
      extractGraphSelection(subsequentClauses).map { case (subsequentGraphSelection, otherClauses) =>
        PartitionedClauses(None, Some(importingWith), Some(subsequentGraphSelection), otherClauses)
      }.getOrElse {
        PartitionedClauses(None, Some(importingWith), None, subsequentClauses)
      }
    }

  private def startingWithGraphSelection(clauses: Seq[Clause]): Option[PartitionedClauses] =
    extractGraphSelection(clauses).map { case (initialGraphSelection, subsequentClauses) =>
      extractImportingWith(subsequentClauses).map { case (importingWith, otherClauses) =>
        PartitionedClauses(Some(initialGraphSelection), Some(importingWith), None, otherClauses)
      }.getOrElse {
        PartitionedClauses(Some(initialGraphSelection), None, None, subsequentClauses)
      }
    }

  private def extractImportingWith(clauses: Seq[Clause]): Option[(With, Seq[Clause])] =
    clauses.headOption.collect {
      case withClause @ With(false, ri, None, None, None, None, _) if ri.items.forall(_.isPassThrough) =>
        (withClause, clauses.tail)
    }

  private def extractGraphSelection(clauses: Seq[Clause]): Option[(UseGraph, Seq[Clause])] =
    clauses.headOption.collect {
      case useGraph: UseGraph => (useGraph, clauses.tail)
    }
}

case object TopLevelBraces extends UnaliasedNotAllowed {
  val name: String = "{ ... }"
  override val msg: String = "{ RETURN ... }"
}

case class TopLevelBraces(
  query: Query,
  use: Option[UseGraph]
)(override val position: InputPosition) extends PartQuery {

  override def singleQuery: SingleQuery = wrapQuery(query, position)
  override def clauses: Seq[Clause] = singleQuery.clauses

  val getCommandClauses: Seq[CommandClause] = query.getCommandClauses

  override def mapEachSingleQuery(f: SingleQuery => SingleQuery): Query = f(wrapQuery(query, position))

  override def getQuery(fromUnion: Boolean): Query = if (fromUnion) {
    wrapQuery(query, position)
  } else {
    query.getQuery(fromUnion) match {
      case u @ UnionDistinct(lhs, rhs) => u.copy(lhs = lhs.getQuery(true), rhs = rhs.singleQuery)(u.position)
      case u @ UnionAll(lhs, rhs)      => u.copy(lhs = lhs.getQuery(true), rhs = rhs.singleQuery)(u.position)
      case wh @ ConditionalQueryWhen(branches, default) =>
        wh.copy(branches = branches.map(_.wrapInnerQuery), default = default.map(_.wrapInnerQuery))(wh.position)
      case q => q
    }
  }

  private def wrapQuery(innerQuery: Query, position: InputPosition): SingleQuery = {
    val lastClause =
      if (innerQuery.isReturning)
        Return(ReturnItems(includeExisting = true, Seq.empty)(position))(position)
          .copy()(position)
      else Finish()(position)

    val sq = SingleQuery(
      Seq(
        ScopeClauseSubqueryCall(
          innerQuery,
          isImportingAll = true,
          Seq.empty,
          None,
          optional = false
        )(position),
        lastClause
      )
    )(position)
    sq
  }

  override def semanticCheckInConditionalQueryContext: SemanticCheck = semanticCheck

  override def semanticCheckInSubqueryContext(outer: SemanticState, current: SemanticState): SemanticCheck =
    query.semanticCheckInSubqueryContext(outer, current) chain recordCurrentScope(this)

  override def semanticCheckImportingWithSubQueryContext(outer: SemanticState): SemanticCheck =
    SemanticCheck.error(SemanticError.invalidUseOfOldCall(TopLevelBraces.name, position))

  override def semanticCheckInSubqueryExpressionContext(
    canOmitReturn: Boolean,
    outer: SemanticState,
    context: UnaliasedNotAllowed
  ): SemanticCheck =
    query.semanticCheckInSubqueryExpressionContext(canOmitReturn, outer, context) chain recordCurrentScope(this)

  override def semanticCheckInTopLevelBracesContext: SemanticCheck = semanticCheck

  override def semanticCheck: SemanticCheck = query.semanticCheckInTopLevelBracesContext chain recordCurrentScope(this)

  override def checkImportingWith: SemanticCheck = query.checkImportingWith
  override def importColumns: Seq[String] = query.importColumns
  override def isCorrelated: Boolean = query.isCorrelated

  override def returnVariables: ReturnVariables = query.returnVariables
}

object Union {

  /**
   * This defines a mapping of variables in both parts of the union to variables valid in the scope after the union.
   */
  case class UnionMapping(
    unionVariable: LogicalVariable,
    variableInLhs: LogicalVariable,
    variableInRhs: LogicalVariable
  )
}

sealed trait Union extends Query {
  def lhs: Query
  def rhs: PartQuery

  def unionMappings: List[UnionMapping]

  override def getQuery(fromUnion: Boolean): Query = this

  val getCommandClauses: Seq[CommandClause] = lhs.getCommandClauses ++ rhs.getCommandClauses

  override def returnVariables: ReturnVariables = ReturnVariables(
    // If either side of the UNION has a RETURN *,
    // then returnVariables.explicitVariables will not list all variables.
    // Instead, one has to inspect `finalScope` to find all variables.
    lhs.returnVariables.includeExisting || rhs.returnVariables.includeExisting,
    unionMappings.map(_.unionVariable)
  )

  override def getReturns: Seq[Return] = {
    lhs.getReturns ++ rhs.getReturns
  }

  override def importColumns: Seq[String] = lhs.importColumns ++ rhs.importColumns

  def containsUpdates: Boolean = lhs.containsUpdates || rhs.containsUpdates

  private def checkRecursively(semanticCheck: Query => SemanticCheck): SemanticCheck = {
    def checkSingleQuery(partQuery: PartQuery): SemanticCheck = withScopedState {
      semanticCheck(partQuery) chain
        checkNoInputDataStreamInsideUnionElement(partQuery) chain
        checkNoCallInTransactionInsideUnionElement(partQuery)
    }

    def checkNestedQuery(query: Query): SemanticCheck =
      query match {
        case partQuery: PartQuery       => checkSingleQuery(partQuery)
        case when: ConditionalQueryWhen => withScopedState(semanticCheck(when))
        case union: Union =>
          withScopedState {
            SemanticCheck.nestedCheck(union.checkRecursively(semanticCheck))
          }
      }

    SemanticCheck.fromState(state => {
      checkUnionAggregation chain
        checkNestedQuery(lhs) chain
        SemanticCheck.fromState(newState =>
          SemanticCheck.setState(newState.recordWorkingGraph(state.workingGraph))
        ) chain
        checkSingleQuery(rhs) chain
        checkColumnNamesAgree chain
        defineUnionVariables chain
        SemanticState.recordCurrentScope(this)
    })
  }

  def semanticCheck: SemanticCheck = checkRecursively(_.semanticCheck)

  override def semanticCheckInSubqueryExpressionContext(
    canOmitReturn: Boolean,
    outer: SemanticState,
    context: UnaliasedNotAllowed
  ): SemanticCheck =
    checkRecursively(
      importValuesFromScope(outer.currentScope.scope) chain
        _.semanticCheckInSubqueryExpressionContext(canOmitReturn, outer, context)
    )

  override def checkImportingWith: SemanticCheck =
    SemanticCheck.nestedCheck(lhs.checkImportingWith) chain
      rhs.checkImportingWith

  override def isCorrelated: Boolean = lhs.isCorrelated || rhs.isCorrelated

  override def isReturning: Boolean = rhs.isReturning // we assume lhs has the same value

  override def endsWithFinish: Boolean = rhs.endsWithFinish || lhs.endsWithFinish

  override def semanticCheckInSubqueryContext(outer: SemanticState, current: SemanticState): SemanticCheck = {
    checkRecursively(innerQuery =>
      importValuesFromScope(outer.currentScope.scope) chain
        innerQuery.semanticCheckInSubqueryContext(outer, current)
    )
  }

  override def semanticCheckImportingWithSubQueryContext(outer: SemanticState): SemanticCheck =
    checkRecursively(_.semanticCheckImportingWithSubQueryContext(outer))

  override def semanticCheckInTopLevelBracesContext: SemanticCheck =
    checkRecursively(_.semanticCheckInTopLevelBracesContext)

  override def semanticCheckInConditionalQueryContext: SemanticCheck =
    checkRecursively(_.semanticCheckInConditionalQueryContext)

  private def defineUnionVariables: SemanticCheck = (state: SemanticState) => {
    var result = SemanticCheckResult.success(state.newChildScope)
    val scopeFromLhs = lhs.finalScope(state.scope(lhs).getOrElse(Scope.empty))
    val scopeFromRhs = rhs.finalScope(state.scope(rhs).getOrElse(Scope.empty))

    /**
     * Derived from UnionMapping, but only has the names of the variables in LHS and RHS,
     * since that is also the information we need here.
     */
    case class Mapping(
      unionVariable: LogicalVariable,
      variableInLhsName: String,
      variableInRhsName: String
    )

    val mappings = {
      // We need a Set since otherwise we would declare variables multiple times,
      // e.g. if they are listed, but there is also a *,
      // or if both branches have a *.
      val builder = Set.newBuilder[Mapping]
      unionMappings.foreach { um =>
        builder.addOne(Mapping(um.unionVariable, um.variableInLhs.name, um.variableInRhs.name))
      }
      // If there is a RETURN * in at least one of the UNION branches,
      // we need to find out what extra variables to include here.
      if (lhs.returnVariables.includeExisting) {
        scopeFromLhs.symbolNames.foreach { name =>
          builder.addOne(Mapping(Variable(name)(this.position, Variable.isIsolatedDefault), name, name))
        }
      }
      if (rhs.returnVariables.includeExisting) {
        scopeFromRhs.symbolNames.foreach { name =>
          builder.addOne(Mapping(Variable(name)(this.position, Variable.isIsolatedDefault), name, name))
        }
      }
      builder.result()
    }

    for {
      mapping <- mappings
      symbolFromLhs <- scopeFromLhs.symbol(mapping.variableInLhsName)
      symbolFromRhs <- scopeFromRhs.symbol(mapping.variableInRhsName)
    } yield {
      val unionType = symbolFromLhs.types.union(symbolFromRhs.types)
      result = result.state.declareVariable(mapping.unionVariable, unionType, unionVariable = true) match {
        case Left(err)        => SemanticCheckResult(result.state, err +: result.errors)
        case Right(nextState) => SemanticCheckResult(nextState, result.errors)
      }
    }

    SemanticCheckResult(result.state.popScope, result.errors)
  }

  override def finalScope(scope: Scope): Scope =
    // Union defines all return variables in its own scope using defineUnionVariables
    scope.children.last

  // Check that columns names agree between both parts of the union
  def checkColumnNamesAgree: SemanticCheck

  private def checkNoInputDataStreamInsideUnionElement(query: PartQuery): SemanticCheck =
    query
      .clauses
      .collectFirst {
        case inputDataStream: InputDataStream => inputDataStream
      }
      .foldSemanticCheck { inputDataStream =>
        error("INPUT DATA STREAM is not supported in UNION queries", inputDataStream.position)
      }

  private def checkUnionAggregation: SemanticCheck = (lhs, this) match {
    case (_: PartQuery, _)                                        => None
    case (_: UnionAll, _: UnionAll)                               => None
    case (_: UnionDistinct, _: UnionDistinct)                     => None
    case (_: ProjectingUnionAll, _: ProjectingUnionAll)           => None
    case (_: ProjectingUnionDistinct, _: ProjectingUnionDistinct) => None
    case _                                                        => Some(SemanticError.invalidUseOfUnion(position))
  }

  private def checkNoCallInTransactionInsideUnionElement(query: PartQuery): SemanticCheck =
    SubqueryCall
      .findTransactionalSubquery(query)
      .foldSemanticCheck { nestedCallInTransactions =>
        error(SemanticError.invalidUseOfUnionAndCIT(nestedCallInTransactions.position))
      }
}

/**
 * UnmappedUnion classes are directly produced by the parser.
 * When we do namespacing, we need to convert them the [[ProjectingUnion]].
 * ProjectingUnion is never produced by the parser.
 *
 * This has two reasons:
 * a) We capture how variables are projected from the two final scopes of the parts of the union to the scope
 *    after the union, before the Namespacer changes the names so that the Variable inside and outside of the union have different names
 *    and we would not find them any longer. The Namespacer will still change the name, but since we captured the Variable and not the
 *    name, we still have the correct projecting information.
 * b) We need to disable `checkColumnNamesAgree` for ProjectingUnion, because the names will actually not agree any more after the namespacing.
 *    This is not a problem though, since we would have failed earlier if the names did not agree originally.
 */
sealed trait UnmappedUnion extends Union {

  // A value instead of a def prevents us from creating new variables every time this is used.
  // This is helpful if the variable is used by reference from the semantic state.
  private var _unionMappings = {
    for {
      lhsCol <- lhs.returnColumns
      rhsCol <- rhs.returnColumns.find(_.name == lhsCol.name)
    } yield {
      // This assumes that lhs.returnColumns and rhs.returnColumns agree
      UnionMapping(Variable(lhsCol.name)(this.position, Variable.isIsolatedDefault), lhsCol, rhsCol)
    }
  }

  override def unionMappings: List[UnionMapping] = _unionMappings

  override def dup(children: Seq[AnyRef]): UnmappedUnion.this.type = {
    val res = super.dup(children)

    val thisLhsCols = lhs.returnColumns
    val thisRhsCols = rhs.returnColumns
    val resLhsCols = res.lhs.returnColumns
    val resRhsCols = res.rhs.returnColumns

    def containTheSameInstances[X <: AnyRef](a: Seq[X], b: Seq[X]): Boolean =
      a.forall(elemA => b.exists(elemB => elemA eq elemB)) && a.size == b.size

    // If we have not rewritten any return column (by reference equality), then we can simply reuse this.unionMappings.
    // This is important because the variables are used by reference from the semantic state.
    if (containTheSameInstances(thisLhsCols, resLhsCols) && containTheSameInstances(thisRhsCols, resRhsCols)) {
      res._unionMappings = this.unionMappings
    }

    res
  }

  override def checkColumnNamesAgree: SemanticCheck = (state: SemanticState) => {
    val lhsScope = if (lhs.isReturning) lhs.finalScope(state.scope(lhs).getOrElse(Scope.empty)) else Scope.empty
    val rhsScope = if (rhs.isReturning) rhs.finalScope(state.scope(rhs).getOrElse(Scope.empty)) else Scope.empty
    val errors =
      if (lhsScope.symbolNames == rhsScope.symbolNames) {
        Seq.empty
      } else {
        Seq(SemanticError.incompatibleReturnColumns(errorParam, position))
      }
    SemanticCheckResult(state, errors)
  }
}

object UnmappedUnion {
  val errorParam = "UNION subqueries"
}

sealed trait ProjectingUnion extends Union {
  // If we have a ProjectingUnion we have already checked this before and now they have been rewritten to actually not match.
  override def checkColumnNamesAgree: SemanticCheck = SemanticCheck.success
}

final case class UnionAll(lhs: Query, rhs: PartQuery)(
  val position: InputPosition
) extends UnmappedUnion {

  override def mapEachSingleQuery(f: SingleQuery => SingleQuery): Query =
    copy(lhs.mapEachSingleQuery(f), f(rhs.singleQuery))(position)
}

final case class UnionDistinct(lhs: Query, rhs: PartQuery)(
  val position: InputPosition
) extends UnmappedUnion {

  override def mapEachSingleQuery(f: SingleQuery => SingleQuery): Query =
    copy(lhs.mapEachSingleQuery(f), f(rhs.singleQuery))(position)
}

final case class ProjectingUnionAll(lhs: Query, rhs: PartQuery, unionMappings: List[UnionMapping])(
  val position: InputPosition
) extends ProjectingUnion {

  override def mapEachSingleQuery(f: SingleQuery => SingleQuery): Query =
    copy(lhs.mapEachSingleQuery(f), f(rhs.singleQuery))(position)
}

final case class ProjectingUnionDistinct(lhs: Query, rhs: PartQuery, unionMappings: List[UnionMapping])(
  val position: InputPosition
) extends ProjectingUnion {

  override def mapEachSingleQuery(f: SingleQuery => SingleQuery): Query =
    copy(lhs.mapEachSingleQuery(f), f(rhs.singleQuery))(position)
}

case class ConditionalQueryBranch(predicate: Expression, query: PartQuery)(val position: InputPosition) extends ASTNode
    with SemanticCheckable with SemanticAnalysisTooling with QueryUtils {

  override def returnVariables: ReturnVariables = query.returnVariables
  override def isReturning: Boolean = query.isReturning

  override def semanticCheck: SemanticCheck =
    semanticCheckAbstract(_.semanticCheckInConditionalQueryContext)

  override def semanticCheckInConditionalQueryContext: SemanticCheck =
    semanticCheck

  override def semanticCheckInTopLevelBracesContext: SemanticCheck =
    semanticCheckAbstract(_.semanticCheckInTopLevelBracesContext)

  override def semanticCheckInSubqueryContext(outer: SemanticState, current: SemanticState): SemanticCheck = {
    importValuesFromScope(outer.currentScope.scope) chain
      semanticCheckAbstract(_.semanticCheckInSubqueryContext(outer, current))
  }

  override def semanticCheckImportingWithSubQueryContext(outer: SemanticState): SemanticCheck =
    semanticCheckAbstract(_.semanticCheckImportingWithSubQueryContext(outer))

  override def semanticCheckInSubqueryExpressionContext(
    canOmitReturn: Boolean,
    outer: SemanticState,
    context: UnaliasedNotAllowed
  ): SemanticCheck =
    semanticCheckAbstract(
      importValuesFromScope(outer.currentScope.scope) chain
        _.semanticCheckInSubqueryExpressionContext(canOmitReturn, outer, context)
    )

  private def semanticCheckAbstract(check: QueryUtils => SemanticCheck): SemanticCheck = {
    predicateCheck chain
      check(query) chain
      recordCurrentScope(this)
  }

  private def predicateCheck: SemanticCheck = {
    SemanticExpressionCheck.simple(predicate) chain
      SemanticExpressionCheck.expectType(CTBoolean.covariant, predicate)
  }

  override def finalScope(scope: Scope): Scope =
    if (scope.children.size < 1) Scope.empty else scope.children.last

  def mapEachSingleQuery(f: SingleQuery => SingleQuery): ConditionalQueryBranch =
    copy(query = f(query.singleQuery))(position)

  def wrapInnerQuery: ConditionalQueryBranch = this.copy(query = query.singleQuery)(position)

}

case object ConditionalQueryWhen extends UnaliasedNotAllowed {
  override val msg: String = "WHEN ... THEN ..."
  val name: String = "conditional queries"
}

case class ConditionalQueryWhen(
  branches: Seq[ConditionalQueryBranch],
  default: Option[ConditionalQueryBranch]
)(val position: InputPosition) extends Query {

  private def allBranches: Seq[ConditionalQueryBranch] = branches ++ default

  override def containsUpdates: Boolean = allBranches.exists(_.query.containsUpdates)

  override def getQuery(fromUnion: Boolean): Query = this

  override def getCommandClauses: Seq[CommandClause] =
    allBranches.flatMap(branch => branch.query.getCommandClauses)

  override def returnVariables: ReturnVariables =
    allBranches.foldLeft(ReturnVariables.empty)((acc, branch) => acc.merge(branch.query.returnVariables))

  override def getReturns: Seq[Return] =
    allBranches.flatMap(_.query.getReturns)

  override def isReturning: Boolean = branches.head.isReturning
  override def endsWithFinish: Boolean = allBranches.exists(_.query.endsWithFinish)

  override def finalScope(scope: Scope): Scope = {
    if (scope.children.size < 1) Scope.empty else scope.children.last
  }

  override def semanticCheckInSubqueryContext(outer: SemanticState, current: SemanticState): SemanticCheck =
    semanticCheckAbstract(_.semanticCheckInSubqueryContext(outer, current))

  override def semanticCheckImportingWithSubQueryContext(outer: SemanticState): SemanticCheck =
    SemanticCheck.error(SemanticError.invalidUseOfOldCall(msg, position))

  override def semanticCheckInSubqueryExpressionContext(
    canOmitReturn: Boolean,
    outer: SemanticState,
    context: UnaliasedNotAllowed
  ): SemanticCheck =
    semanticCheckAbstract(_.semanticCheckInSubqueryExpressionContext(canOmitReturn, outer, ConditionalQueryWhen))

  override def semanticCheckInTopLevelBracesContext: SemanticCheck =
    semanticCheckAbstract(_.semanticCheckInTopLevelBracesContext)

  override def semanticCheckInConditionalQueryContext: SemanticCheck = semanticCheck

  override def semanticCheck: SemanticCheck =
    semanticCheckAbstract(_.semanticCheckInConditionalQueryContext)

  private def semanticCheckAbstract(check: QueryUtils => SemanticCheck): SemanticCheck = {
    allBranches.foldSemanticCheck(x => withScopedState(check(x))) chain
      checkReturns chain
      defineReturnScope chain
      recordCurrentScope(this)
  }

  private def defineReturnScope: SemanticCheck = (state: SemanticState) => {
    val result = SemanticCheckResult.success(state.newChildScope)
    val headQuery = branches.head.query
    val headScope = state.scope(headQuery)
    val scope = headQuery.finalScope(headScope.getOrElse(Scope.empty))
    SemanticCheckResult(result.state.importValuesFromScope(scope).popScope, Seq.empty)
  }

  // Return checks
  // 1. All conditional queries need to be either returning or unit
  // 2. There need to be the same number of return values
  // 3. Column names need to agree

  private def firstNonConformer(f: QueryUtils => Any): Option[PartQuery] = {
    val firstQueryType = f(branches.head)
    allBranches.find(f(_) != firstQueryType).map(_.query)
  }

  private def checkConformingBranches: SemanticCheck =
    when(firstNonConformer(_.isReturning).isDefined)((state: SemanticState) => {
      SemanticCheckResult(
        state,
        Seq(SemanticError.incompatibleSubqueryType(name, firstNonConformer(_.isReturning).get.position))
      )
    })

  private def checkColumnNamesAgree: SemanticCheck = (state: SemanticState) => {
    val myScope: Scope = state.currentScope.scope

    val scopes = allBranches.zipWithIndex.map { case (b, i) =>
      (b, if (b.query.isReturning) b.finalScope(myScope.children(i)) else Scope.empty)
    }

    val errors = scopes.foldLeft(Seq.empty[SemanticError]) { case (errors, (branch, scope)) =>
      if (scopes.head._2.symbolNames.size != scope.symbolNames.size)
        errors ++ Seq(SemanticError.incompatibleNumberOfReturnColumns(name, branch.position))
      else if (scopes.head._2.symbolNames != scope.symbolNames)
        errors ++ Seq(SemanticError.incompatibleWhenReturnColumns(name, branch.position))
      else errors
    }
    SemanticCheckResult(state, errors)
  }

  private def checkReturns: SemanticCheck =
    checkConformingBranches ifOkChain
      checkColumnNamesAgree

  override def checkImportingWith: SemanticCheck =
    allBranches.foldSemanticCheck(_.query.checkImportingWith)

  override def isCorrelated: Boolean =
    allBranches.exists(_.query.isCorrelated)

  override def importColumns: Seq[String] =
    allBranches.flatMap(_.query.importColumns)

  override def mapEachSingleQuery(f: SingleQuery => SingleQuery): Query =
    copy(branches.map(_.mapEachSingleQuery(f)), default.map(_.mapEachSingleQuery(f)))(position)
}
