/*
 * Copyright (c) "Neo4j"
 * Neo4j Sweden AB [https://neo4j.com]
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.neo4j.cypher.internal.frontend

import org.neo4j.cypher.internal.CypherVersion
import org.neo4j.cypher.internal.ast.Ast.p
import org.neo4j.cypher.internal.ast.semantics.SemanticError
import org.neo4j.cypher.internal.ast.semantics.SemanticFeature.MatchModes
import org.neo4j.cypher.internal.frontend.helpers.ShortestSyntax
import org.neo4j.cypher.internal.util.InputPosition
import org.neo4j.cypher.internal.util.test_helpers.CypherFunSuite
import org.neo4j.cypher.internal.util.test_helpers.TestName
import org.neo4j.gqlstatus.GqlHelper

class MatchModesSemanticAnalysisTest extends CypherFunSuite
    with SemanticAnalysisTestSuiteWithDefaultQuery
    with TestName
    with ShortestSyntax {

  override def defaultQuery: String = s"MATCH $testName RETURN *"

  def unboundRepeatableElementsSemanticError(pos: InputPosition): SemanticError = SemanticError(
    GqlHelper.getGql42001_42N53(pos.offset, pos.line, pos.column),
    "The quantified path pattern may yield an infinite number of rows under match mode 'REPEATABLE ELEMENTS'. " +
      "Add an upper bound to the quantified path pattern.",
    pos
  )

  def differentRelationshipsSelectivePathPatternSemanticError(
    pos: InputPosition,
    semanticFeatureEnabled: Boolean = true
  ): SemanticError = {
    val matchModeTip = if (semanticFeatureEnabled) {
      " You may want to use multiple MATCH clauses, or you might want to consider using the REPEATABLE ELEMENTS match mode."
    } else {
      ""
    }

    SemanticError(
      GqlHelper.getGql42001_42I45(matchModeTip, pos.offset, pos.line, pos.column),
      "Multiple path patterns cannot be used in the same clause in combination with a selective path selector." + matchModeTip,
      pos
    )
  }

  test("REPEATABLE ELEMENTS (c5)") {
    // explicit match mode is not supported in Cypher 5
    run(disabledCypherVersions = Set(CypherVersion.Cypher25)).hasError(
      GqlHelper.getGql42001_42N54("REPEATABLE ELEMENTS", 6, 1, 7),
      "Match modes such as `REPEATABLE ELEMENTS` are not supported in Cypher 5.",
      p(6, 1, 7)
    )
  }

  test("DIFFERENT RELATIONSHIPS (c5)") {
    // explicit match mode is not supported in Cypher 5
    run(disabledCypherVersions = Set(CypherVersion.Cypher25)).hasError(
      GqlHelper.getGql42001_42N54("DIFFERENT RELATIONSHIPS", 6, 1, 7),
      "Match modes such as `DIFFERENT RELATIONSHIPS` are not supported in Cypher 5.",
      p(6, 1, 7)
    )
  }

  test("REPEATABLE ELEMENTS (c25)") {
    // running without semantic feature should fail
    run(disabledCypherVersions = Set(CypherVersion.Cypher5)).hasError(
      GqlHelper.getGql42001_42N54("REPEATABLE ELEMENTS", 6, 1, 7),
      "Match modes such as `REPEATABLE ELEMENTS` are not supported yet.",
      p(6, 1, 7)
    )
  }

  test("DIFFERENT RELATIONSHIPS (c25)") {
    // running without semantic feature should fail
    run(disabledCypherVersions = Set(CypherVersion.Cypher5)).hasError(
      GqlHelper.getGql42001_42N54("DIFFERENT RELATIONSHIPS", 6, 1, 7),
      "Match modes such as `DIFFERENT RELATIONSHIPS` are not supported yet.",
      p(6, 1, 7)
    )
  }

  test("(a)") {
    // running with implicit "DIFFERENT RELATIONSHIPS" match mode should not fail
    run().hasNoErrors
  }

  test("DIFFERENT RELATIONSHIPS ((a)-[:REL]->(b)){2}") {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasNoErrors
  }

  test("((a)-[:REL]->(b)){2}, ((c)-[:REL]->(d))+") {
    runWith(MatchModes).hasNoErrors
  }

  test("REPEATABLE ELEMENTS ((a)-[:REL]->(b)){2}") {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasNoErrors
  }

  test("REPEATABLE ELEMENTS ((a)-[:REL]->(b)){1,}") {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasErrors(
      unboundRepeatableElementsSemanticError(p(26, 1, 27))
    )
  }

  test("REPEATABLE ELEMENTS ((a)-[:REL]->(b))+") {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasErrors(
      unboundRepeatableElementsSemanticError(p(26, 1, 27))
    )
  }

  test("REPEATABLE ELEMENTS (a)-[:REL*]->(b)") {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasErrors(
      unboundRepeatableElementsSemanticError(p(26, 1, 27))
    )
  }

  test("REPEATABLE ELEMENTS SHORTEST 2 PATH ((a)-[:REL]->(b))+") {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasErrors(
      unboundRepeatableElementsSemanticError(p(42, 1, 43))
    )
  }

  test("REPEATABLE ELEMENTS ANY ((a)-[:REL]->(b))+") {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasErrors(
      unboundRepeatableElementsSemanticError(p(30, 1, 31))
    )
  }

  test("REPEATABLE ELEMENTS SHORTEST 1 PATH GROUPS ((a)-[:REL]->(b))+") {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasErrors(
      unboundRepeatableElementsSemanticError(p(49, 1, 50))
    )
  }

  test("shortestPath((a)-[:REL*]->(b)), shortestPath((c)-[:REL*]->(d))") {
    runWith(MatchModes).hasNoErrors
  }

  test("REPEATABLE ELEMENTS ((a)-[:REL]->(b)){2}, ((c)-[:REL]->(d))+") {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasErrors(
      unboundRepeatableElementsSemanticError(p(48, 1, 49))
    )
  }

  test("REPEATABLE ELEMENTS ((a)-[:REL]->(b))+, ((c)-[:REL]->(d))+") {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasErrors(
      unboundRepeatableElementsSemanticError(p(26, 1, 27)),
      unboundRepeatableElementsSemanticError(p(46, 1, 47))
    )
  }

  test("REPEATABLE ELEMENTS SHORTEST 1 PATH (a)-[:REL*]->(b), SHORTEST 1 PATH (c)-[:REL*]->(d)") {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasErrors(
      unboundRepeatableElementsSemanticError(p(42, 1, 43)),
      unboundRepeatableElementsSemanticError(p(76, 1, 77))
    )
  }

  test("DIFFERENT RELATIONSHIPS SHORTEST 1 PATH (a)-[:REL*]->(b), SHORTEST 1 PATH (c)-[:REL*]->(d)") {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasErrors(
      differentRelationshipsSelectivePathPatternSemanticError(p(46, 1, 47))
    )
  }

  test("SHORTEST 1 PATH (a)-[:REL*]->(b), (c)-[:REL]->(d)") {
    runWith(MatchModes).hasErrors(
      differentRelationshipsSelectivePathPatternSemanticError(p(22, 1, 23))
    )
  }

  test("SHORTEST 1 PATH (a)-[:REL*]->(b), (c)-[:REL]->(e)") {
    // running without MatchModes
    run().hasErrors(
      differentRelationshipsSelectivePathPatternSemanticError(p(22, 1, 23), semanticFeatureEnabled = false)
    )
  }

  private def invalidEquijoinErrorMsg(varName: String) = s"Variable `$varName` already declared"

  // Under REPEATABLE ELEMENTS, we allow mixing selective path patterns alongside other path patterns
  allSelectiveSelectors.foreach { selector =>
    allSelectors.foreach { otherSelector =>
      test(
        s"""MATCH REPEATABLE ELEMENTS
           |   p1 = $selector (a)-->{1, 32}(b)-->(c),
           |   p2 = $otherSelector (x)-->{1, 32}(y)-->(z)
           |RETURN count(*)""".stripMargin
      ) {
        runWith(testName, disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasNoErrors
      }
      test(
        s"""MATCH REPEATABLE ELEMENTS
           |   p2 = $otherSelector (x)-->{1, 32}(y)-->(z),
           |   p1 = $selector (a)-->{1, 32}(b)-->(c)
           |RETURN count(*)""".stripMargin
      ) {
        runWith(testName, disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasNoErrors
      }

      // Newly introduced strict interior nodes of selective path patterns may not be reused in the same MATCH clause
      test(
        s"""MATCH REPEATABLE ELEMENTS
           |   p1 = $selector (a)-->{1, 32}(x)-->(c),
           |   p2 = $otherSelector (x)-->{1, 32}(y)-->(z)
           |RETURN count(*)""".stripMargin
      ) {
        runWith(
          testName,
          disabledCypherVersions = Set(CypherVersion.Cypher5),
          MatchModes
        ).hasErrorMessages(invalidEquijoinErrorMsg("x"))
      }
      test(
        s"""MATCH REPEATABLE ELEMENTS
           |   p2 = $otherSelector (x)-->{1, 32}(y)-->(z),
           |   p1 = $selector (a)-->{1, 32}(x)-->(c)
           |RETURN count(*)""".stripMargin
      ) {
        runWith(
          testName,
          disabledCypherVersions = Set(CypherVersion.Cypher5),
          MatchModes
        ).hasErrorMessages(invalidEquijoinErrorMsg("x"))
      }
      // when the variable is already bound in the previous match clause
      test(
        s"""MATCH (x)
           |MATCH REPEATABLE ELEMENTS
           |   p1 = $selector (a)-->{1, 32}(x)-->(c),
           |   p2 = $otherSelector (x)-->{1, 32}(y)-->(z)
           |RETURN count(*)""".stripMargin
      ) {
        runWith(testName, disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasNoErrors
      }
      test(
        s"""MATCH (x)
           |MATCH REPEATABLE ELEMENTS
           |   p2 = $otherSelector (x)-->{1, 32}(y)-->(z),
           |   p1 = $selector (a)-->{1, 32}(x)-->(c)
           |RETURN count(*)""".stripMargin
      ) {
        runWith(testName, disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasNoErrors
      }
    }
    allSelectiveSelectors.foreach { otherSelector =>
      // overlap between strict interior nodes of different SPPs is not okay either
      test(
        s"""MATCH REPEATABLE ELEMENTS
           |   p2 = $otherSelector (x)-->{1, 32}(y)-->(z),
           |   p1 = $selector (a)-->{1, 32}(y)-->(c)
           |RETURN count(*)""".stripMargin
      ) {
        runWith(
          testName,
          disabledCypherVersions = Set(CypherVersion.Cypher5),
          MatchModes
        ).hasErrorMessages(invalidEquijoinErrorMsg("y"))
      }

      test(
        s"""MATCH REPEATABLE ELEMENTS
           |   p2 = $otherSelector (x)-->{1, 32}(y)-->(z),
           |   p1 = $selector (a)-->{1, 32}(y)-->(y)
           |RETURN count(*)""".stripMargin
      ) {
        runWith(
          testName,
          disabledCypherVersions = Set(CypherVersion.Cypher5),
          MatchModes
        ).hasErrorMessages(invalidEquijoinErrorMsg("y"))
      }

      // but it is if the nodes are boundary nodes as well
      test(
        s"""MATCH REPEATABLE ELEMENTS
           |   p2 = $otherSelector (x)-->{1, 32}(y)-->(y),
           |   p1 = $selector (a)-->{1, 32}(y)-->(y)
           |RETURN count(*)""".stripMargin
      ) {
        runWith(testName, disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasNoErrors
      }

      // Or if they are previously bound
      test(
        s"""MATCH (y)
           |MATCH REPEATABLE ELEMENTS
           |   p2 = $otherSelector (x)-->{1, 32}(y)-->(z),
           |   p1 = $selector (a)-->{1, 32}(y)-->(c)
           |RETURN count(*)""".stripMargin
      ) {
        runWith(testName, disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasNoErrors
      }
      test(
        s"""MATCH (y)
           |MATCH REPEATABLE ELEMENTS
           |   p2 = $otherSelector (x)-->{1, 32}(y)-->(z),
           |   p1 = $selector (a)-->{1, 32}(y)-->(y)
           |RETURN count(*)""".stripMargin
      ) {
        runWith(testName, disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasNoErrors
      }

    }
  }

  test("(a)-[:REL]->(b), ALL PATHS (c)-[:REL]->(d)") {
    runWith(MatchModes).hasNoErrors
  }

  test("REPEATABLE ELEMENTS (a)-[:REL]->(b), ALL PATHS (c)-[:REL]->(d)") {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasNoErrors
  }

  test("REPEATABLE ELEMENTS shortestPath((a)-->(b))") {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasError(
      GqlHelper.getGql42001_42I39("shortestPath", 26, 1, 27),
      "Mixing shortestPath/allShortestPaths with path selectors (e.g. 'ANY SHORTEST') or explicit match modes ('e.g. DIFFERENT RELATIONSHIPS') is not allowed.",
      p(26, 1, 27)
    )
  }

  test("DIFFERENT RELATIONSHIPS shortestPath((a)-->(b))") {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasError(
      GqlHelper.getGql42001_42I39("shortestPath", 30, 1, 31),
      "Mixing shortestPath/allShortestPaths with path selectors (e.g. 'ANY SHORTEST') or explicit match modes ('e.g. DIFFERENT RELATIONSHIPS') is not allowed.",
      p(30, 1, 31)
    )
  }

  test("REPEATABLE ELEMENTS allShortestPaths((a)-->(b))") {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasError(
      GqlHelper.getGql42001_42I39("allShortestPaths", 26, 1, 27),
      "Mixing shortestPath/allShortestPaths with path selectors (e.g. 'ANY SHORTEST') or explicit match modes ('e.g. DIFFERENT RELATIONSHIPS') is not allowed.",
      p(26, 1, 27)
    )
  }

  test("REPEATABLE ELEMENTS (a)-->(b) WHERE shortestPath((a)-->(b)) IS NOT NULL") {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasError(
      GqlHelper.getGql42001_42I39("shortestPath", 42, 1, 43),
      "Mixing shortestPath/allShortestPaths with path selectors (e.g. 'ANY SHORTEST') or explicit match modes ('e.g. DIFFERENT RELATIONSHIPS') is not allowed.",
      p(42, 1, 43)
    )
  }

  test("REPEATABLE ELEMENTS (a)-->(b) WHERE EXISTS { MATCH shortestPath((a)-->(b)) }") {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasError(
      GqlHelper.getGql42001_42I39("shortestPath", 57, 1, 58),
      "Mixing shortestPath/allShortestPaths with path selectors (e.g. 'ANY SHORTEST') or explicit match modes ('e.g. DIFFERENT RELATIONSHIPS') is not allowed.",
      p(57, 1, 58)
    )
  }

  test("CALL { MATCH REPEATABLE ELEMENTS (a)-->(b) MATCH shortestPath((c)-->(d)) RETURN * } RETURN *") {
    run(testName, pipelineWithSemanticFeatures(MatchModes), disabledVersions = Set(CypherVersion.Cypher5)).hasError(
      GqlHelper.getGql42001_42I39("shortestPath", 49, 1, 50),
      "Mixing shortestPath/allShortestPaths with path selectors (e.g. 'ANY SHORTEST') or explicit match modes ('e.g. DIFFERENT RELATIONSHIPS') is not allowed.",
      p(49, 1, 50)
    )
  }

  test(s"REPEATABLE ELEMENTS (a)-->(b) MATCH shortestPath((c)-->(d))") {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasNoErrors
  }

  test(s"DIFFERENT RELATIONSHIPS (a)-->(b) MATCH shortestPath((c)-->(d))") {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasNoErrors
  }

  test(
    "REPEATABLE ELEMENTS ALL SHORTEST (a)((n1)-[r1]->(n2)){1,5}(b), ALL SHORTEST (b)-[r2]->{2,3}(c)"
  ) {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasNoErrors
  }

  test(
    "REPEATABLE ELEMENTS ALL SHORTEST (a)((n1)-[r1:R1]->(n2)){1,5}(b)-[r2:R2]->{2,3}(c)"
  ) {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasNoErrors
  }

  test(
    "REPEATABLE ELEMENTS (b)-[r3:R1]->(c),  ALL SHORTEST (a)((n1)-[r1:R1]->(n2)){1,5}(b)-[r2:R2]->{2,3}(b)"
  ) {
    // b is a strict interior node, but also a boundary node. Therefore, it can be used in other path-patterns.
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasNoErrors
  }

  test(
    "REPEATABLE ELEMENTS ALL SHORTEST (a)((n1)-[r1:R1]->(n2)){1,2}(i)-[r2:R2]->{2,3}(b), (b)-[r3:R1]->(c), SHORTEST 1 (c)-[:R2]->{2,3}(d)"
  ) {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasNoErrors
  }

  test(
    "REPEATABLE ELEMENTS (a)--(b), ALL SHORTEST ((b)-[r2]->{2,3}(c) WHERE b.p=a.p)"
  ) {
    // selective path pattern has a predicate referring to another path pattern in the same graph pattern - reference to variable in earlier path pattern
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasError(
      GqlHelper.getGql42001_42I21(
        java.util.List.of("a"),
        "ALL SHORTEST PATHS ((b) (()-[r2]->()){2, 3} (c) WHERE b.p = a.p)",
        79,
        1,
        80
      ),
      """From within a selective path pattern, one may only reference variables, that are already bound in a previous `MATCH` clause.
        |In this case, `a` is defined in the same `MATCH` clause as ALL SHORTEST PATHS ((b) (()-[r2]->()){2, 3} (c) WHERE b.p = a.p).""".stripMargin,
      p(79, 1, 80)
    )
  }

  test(
    "REPEATABLE ELEMENTS ALL SHORTEST ((b)-[r2]->{2,3}(c) WHERE b.p=a.p), (a)--{,2}(b)"
  ) {
    // selective path pattern has a predicate referring to another path pattern in the same graph pattern - reference to variable in later path pattern
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasError(
      GqlHelper.getGql42001_42I21(
        java.util.List.of("a"),
        "ALL SHORTEST PATHS ((b) (()-[r2]->()){2, 3} (c) WHERE b.p = a.p)",
        69,
        1,
        70
      ),
      """From within a selective path pattern, one may only reference variables, that are already bound in a previous `MATCH` clause.
        |In this case, `a` is defined in the same `MATCH` clause as ALL SHORTEST PATHS ((b) (()-[r2]->()){2, 3} (c) WHERE b.p = a.p).""".stripMargin,
      p(69, 1, 70)
    )
  }

  test(
    "(a)--(b) MATCH REPEATABLE ELEMENTS ALL SHORTEST ((b)-[r2]->{2,3}(c) WHERE a.p=c.p)"
  ) {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasNoErrors
  }

  test(
    "ALL SHORTEST (x)--{,2}(y)--{,2}(z) MATCH REPEATABLE ELEMENTS ALL SHORTEST ((b)-[r2]->{2,3}(c) WHERE b.p=y.p)"
  ) {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasNoErrors
  }

  test(
    "(a)-[:R]->(e) MATCH REPEATABLE ELEMENTS (a)--(b), ALL SHORTEST ((b)--{2,3}(c)--{,2}(d) WHERE c.p=a.p)"
  ) {
    // a is already bound in a previous clause, therefore c.p=a.p is allowed here
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasNoErrors
  }

  test(
    "(a)-[:R]->(e) MATCH REPEATABLE ELEMENTS (a)--(b), ALL SHORTEST ((b)--{2,3}(c) WHERE c.p=a.p)"
  ) {
    runWith(disabledCypherVersions = Set(CypherVersion.Cypher5), MatchModes).hasNoErrors
  }
}
