/*
 * Copyright (c) "Neo4j"
 * Neo4j Sweden AB [https://neo4j.com]
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.neo4j.cypher.internal.frontend

import org.neo4j.cypher.internal.CypherVersion.Cypher25
import org.neo4j.cypher.internal.CypherVersion.Cypher5
import org.neo4j.cypher.internal.ast.Ast.p
import org.neo4j.cypher.internal.ast.SingleQuery
import org.neo4j.cypher.internal.ast.semantics.SemanticError
import org.neo4j.cypher.internal.ast.semantics.SemanticFeature
import org.neo4j.cypher.internal.frontend.SemanticAnalysisTest.gql42NA5
import org.neo4j.cypher.internal.util.ErrorMessageProvider
import org.neo4j.cypher.internal.util.SubqueryVariableShadowing
import org.neo4j.cypher.internal.util.test_helpers.CypherFunSuite
import org.neo4j.gqlstatus.ErrorGqlStatusObjectImplementation
import org.neo4j.gqlstatus.GqlHelper
import org.neo4j.gqlstatus.GqlHelper.getGql42001_42N07
import org.neo4j.gqlstatus.GqlHelper.getGql42001_42N71
import org.neo4j.gqlstatus.GqlParams
import org.neo4j.gqlstatus.GqlStatusInfoCodes

class SubqueryCallSemanticAnalysisTest
    extends CypherFunSuite
    with NameBasedSemanticAnalysisTestSuite {

  private val withMultiGraphs = pipelineWithSemanticFeatures(
    SemanticFeature.MultipleGraphs,
    SemanticFeature.UseAsMultipleGraphsSelector
  )

  private val withSingleGraph = pipelineWithSemanticFeatures(
    SemanticFeature.MultipleGraphs,
    SemanticFeature.UseAsSingleGraphSelector
  )

  test("Returning a variable that is already bound outside should give a useful error with scope") {
    val query =
      """WITH 1 AS i
        |CALL () {
        |  WITH 2 AS i
        |  RETURN i
        |}
        |RETURN i
        |""".stripMargin

    run(query).hasError("Variable `i` already declared in outer scope", p(45, 4, 10))
  }

  test("Returning a variable that is already bound outside should give a useful error") {
    val query =
      """WITH 1 AS i
        |CALL {
        |  WITH 2 AS i
        |  RETURN i
        |}
        |RETURN i
        |""".stripMargin

    run(query)
      .hasError(getGql42001_42N07("i", 42, 4, 10), "Variable `i` already declared in outer scope", p(42, 4, 10))
  }

  test("Returning a variable that is already bound outside, from a union, should give a useful error") {
    val query =
      """WITH 1 AS i
        |CALL {
        |  WITH 2 AS i
        |  RETURN i
        |    UNION
        |  WITH 3 AS i
        |  RETURN 2 AS i
        |}
        |RETURN i
        |""".stripMargin

    run(query).hasErrors(
      getGql42001_42N07("i", 42, 4, 10),
      "Variable `i` already declared in outer scope",
      p(42, 4, 10),
      getGql42001_42N07("i", 82, 7, 15),
      "Variable `i` already declared in outer scope",
      p(82, 7, 15)
    )
  }

  test("Returning a variable implicitly that is already bound outside scoped call should give a useful error") {
    val query =
      """WITH 1 AS i
        |CALL () {
        |  WITH 2 AS i
        |  RETURN *
        |}
        |RETURN i
        |""".stripMargin

    run(query).hasError("Variable `i` already declared in outer scope", p(38, 4, 3))
  }

  test("Returning a variable implicitly that is already bound outside should give a useful error") {
    val query =
      """WITH 1 AS i
        |CALL {
        |  WITH 2 AS i
        |  RETURN *
        |}
        |RETURN i
        |""".stripMargin

    run(query).hasError(getGql42001_42N07("i", 42, 4, 10), "Variable `i` already declared in outer scope", p(42, 4, 10))
  }

  test("Returning a variable implicitly that is already bound outside, from a union, should give a useful error") {
    val query =
      """WITH 1 AS i
        |CALL {
        |  WITH 2 AS i
        |  RETURN *
        |    UNION
        |  WITH 3 AS i
        |  RETURN *
        |}
        |RETURN i
        |""".stripMargin

    run(query).hasErrors(
      getGql42001_42N07("i", 42, 4, 10),
      "Variable `i` already declared in outer scope",
      p(42, 4, 10),
      getGql42001_42N07("i", 77, 7, 10),
      "Variable `i` already declared in outer scope",
      p(77, 7, 10)
    )
  }

  test("Should warn about variable shadowing in a scoped subquery") {
    val query =
      """MATCH (shadowed)
        |CALL () {
        |  MATCH (shadowed)-[:REL]->(m) // warning here
        |  RETURN m
        |}
        |RETURN *""".stripMargin
    run(query).hasNotifications(SubqueryVariableShadowing(p(36, 3, 10), "shadowed"))
  }

  test("Should warn about variable shadowing in a subquery") {
    val query =
      """MATCH (shadowed)
        |CALL {
        |  MATCH (shadowed)-[:REL]->(m) // warning here
        |  RETURN m
        |}
        |RETURN *""".stripMargin
    run(query).hasNotifications(SubqueryVariableShadowing(p(33, 3, 10), "shadowed"))
  }

  test("Should warn about variable shadowing in a scoped subquery when aliasing") {
    val query =
      """MATCH (shadowed)
        |CALL () {
        |  MATCH (n)-[:REL]->(m)
        |  WITH m AS shadowed // warning here
        |  WITH shadowed AS m
        |  RETURN m
        |}
        |RETURN *""".stripMargin
    run(query).hasNotifications(SubqueryVariableShadowing(p(63, 4, 13), "shadowed"))
  }

  test("Should warn about variable shadowing in a subquery when aliasing") {
    val query =
      """MATCH (shadowed)
        |CALL {
        |  MATCH (n)-[:REL]->(m)
        |  WITH m AS shadowed // warning here
        |  WITH shadowed AS m
        |  RETURN m
        |}
        |RETURN *""".stripMargin
    run(query).hasNotifications(SubqueryVariableShadowing(p(60, 4, 13), "shadowed"))
  }

  test("Should warn about variable shadowing in a nested scoped subquery") {
    val query =
      """MATCH (shadowed)
        |CALL () {
        |  MATCH (n)-[:REL]->(m)
        |  CALL () {
        |    MATCH (shadowed)-[:REL]->(x) // warning here
        |    RETURN x
        |  }
        |  RETURN m, x
        |}
        |RETURN *""".stripMargin
    run(query).hasNotifications(SubqueryVariableShadowing(p(74, 5, 12), "shadowed"))
  }

  test("Should warn about variable shadowing in a nested subquery") {
    val query =
      """MATCH (shadowed)
        |CALL {
        |  MATCH (n)-[:REL]->(m)
        |  CALL {
        |    MATCH (shadowed)-[:REL]->(x) // warning here
        |    RETURN x
        |  }
        |  RETURN m, x
        |}
        |RETURN *""".stripMargin
    run(query).hasNotifications(SubqueryVariableShadowing(p(68, 5, 12), "shadowed"))
  }

  test("Should warn about variable shadowing from enclosing scoped subquery") {
    val query =
      """MATCH (shadowed)
        |CALL (shadowed) {
        |  MATCH (shadowed)-[:REL]->(m)
        |  CALL () {
        |    MATCH (shadowed)-[:REL]->(x) // warning here
        |    RETURN x
        |  }
        |  RETURN m, x
        |}
        |RETURN *""".stripMargin
    run(query).hasNotifications(SubqueryVariableShadowing(p(89, 5, 12), "shadowed"))
  }

  test("Should warn about variable shadowing from enclosing subquery") {
    val query =
      """MATCH (shadowed)
        |CALL {
        |  WITH shadowed
        |  MATCH (shadowed)-[:REL]->(m)
        |  CALL {
        |    MATCH (shadowed)-[:REL]->(x) // warning here
        |    RETURN x
        |  }
        |  RETURN m, x
        |}
        |RETURN *""".stripMargin
    run(query).hasNotifications(SubqueryVariableShadowing(p(91, 6, 12), "shadowed"))
  }

  test("Should warn about multiple shadowed variables in a scoped subquery") {
    val query =
      """MATCH (shadowed)-->(alsoShadowed)
        |CALL () {
        |  MATCH (shadowed)-->(alsoShadowed) // multiple warnings here
        |  RETURN shadowed AS n, alsoShadowed AS m
        |}
        |RETURN *""".stripMargin
    run(query).hasNotifications(
      SubqueryVariableShadowing(p(53, 3, 10), "shadowed"),
      SubqueryVariableShadowing(p(66, 3, 23), "alsoShadowed")
    )
  }

  test("Should warn about multiple shadowed variables in a subquery") {
    val query =
      """MATCH (shadowed)-->(alsoShadowed)
        |CALL {
        |  MATCH (shadowed)-->(alsoShadowed) // multiple warnings here
        |  RETURN shadowed AS n, alsoShadowed AS m
        |}
        |RETURN *""".stripMargin
    run(query).hasNotifications(
      SubqueryVariableShadowing(p(50, 3, 10), "shadowed"),
      SubqueryVariableShadowing(p(63, 3, 23), "alsoShadowed")
    )
  }

  test("Should warn about multiple shadowed variables in a nested scoped subquery") {
    val query =
      """MATCH (shadowed)
        |CALL () {
        |  MATCH (shadowed)-[:REL]->(m) // warning here
        |  CALL () {
        |    MATCH (shadowed)-[:REL]->(x) // and also here
        |    RETURN x
        |  }
        |  RETURN m, x
        |}
        |RETURN *""".stripMargin
    run(query).hasNotifications(
      SubqueryVariableShadowing(p(36, 3, 10), "shadowed"),
      SubqueryVariableShadowing(p(97, 5, 12), "shadowed")
    )
  }

  test("Should warn about multiple shadowed variables in a nested subquery") {
    val query =
      """MATCH (shadowed)
        |CALL {
        |  MATCH (shadowed)-[:REL]->(m) // warning here
        |  CALL {
        |    MATCH (shadowed)-[:REL]->(x) // and also here
        |    RETURN x
        |  }
        |  RETURN m, x
        |}
        |RETURN *""".stripMargin
    run(query).hasNotifications(
      SubqueryVariableShadowing(p(33, 3, 10), "shadowed"),
      SubqueryVariableShadowing(p(91, 5, 12), "shadowed")
    )
  }

  test(
    "Should not warn about variable shadowing in a subquery if it has been removed from scope by WITH - scoped subquery"
  ) {
    val query =
      """MATCH (notShadowed)
        |WITH notShadowed AS n
        |CALL () {
        |  MATCH (notShadowed)-[:REL]->(m)
        |  RETURN m
        |}
        |RETURN *""".stripMargin
    run(query).hasNotifications()
  }

  test("Should not warn about variable shadowing in a subquery if it has been removed from scope by WITH") {
    val query =
      """MATCH (notShadowed)
        |WITH notShadowed AS n
        |CALL {
        |  MATCH (notShadowed)-[:REL]->(m)
        |  RETURN m
        |}
        |RETURN *""".stripMargin
    run(query).hasNotificationsIn {
      case Cypher25 => Seq.empty
      case Cypher5  => Seq.empty
    }
  }

  test("Should not allow redeclaration of imported variable in a scoped subquery") {
    val query =
      """MATCH (notShadowed)
        |CALL (notShadowed) {
        |  MATCH (notShadowed)-[:REL]->(m)
        |  WITH m AS notShadowed
        |  RETURN notShadowed AS x
        |}
        |RETURN *""".stripMargin
    run(query).hasError(
      "The variable `notShadowed` is shadowing an imported variable with the same name and needs to be renamed",
      p(87, 4, 13)
    )
  }

  test("Should not warn about variable shadowing in a subquery if it has been imported previously") {
    val query =
      """MATCH (notShadowed)
        |CALL {
        |  WITH notShadowed
        |  MATCH (notShadowed)-[:REL]->(m)
        |  WITH m AS notShadowed
        |  RETURN notShadowed AS x
        |}
        |RETURN *""".stripMargin
    run(query).hasNoNotifications
  }

  test("Should warn about variable shadowing in an union scoped subquery") {
    val query =
      """MATCH (shadowed)
        |CALL () {
        |  MATCH (m) RETURN m
        | UNION
        |  MATCH (shadowed)-[:REL]->(m) // warning here
        |  RETURN m
        |}
        |RETURN *""".stripMargin
    run(query).hasNotifications(SubqueryVariableShadowing(p(64, 5, 10), "shadowed"))
  }

  test("Should warn about variable shadowing in an union subquery") {
    val query =
      """MATCH (shadowed)
        |CALL {
        |  MATCH (m) RETURN m
        | UNION
        |  MATCH (shadowed)-[:REL]->(m) // warning here
        |  RETURN m
        |}
        |RETURN *""".stripMargin
    run(query).hasNotifications(SubqueryVariableShadowing(p(61, 5, 10), "shadowed"))
  }

  test("Should warn about variable shadowing in one of the union subquery branches") {
    val query =
      """MATCH (shadowed)
        |CALL {
        |  WITH shadowed
        |  MATCH (shadowed)-[:REL]->(m)
        |  RETURN m
        | UNION
        |  MATCH (shadowed)-[:REL]->(m) // warning here
        |  RETURN m
        | UNION
        |  MATCH (x) RETURN x AS m
        |}
        |RETURN *""".stripMargin
    run(query).hasNotifications(SubqueryVariableShadowing(p(98, 7, 10), "shadowed"))
  }

  test("Subquery with only importing WITH") {
    run("WITH 1 AS a CALL { WITH a } RETURN a").hasError(
      getGql42001_42N71(19, 1, 20),
      "Query must conclude with a RETURN clause, a FINISH clause, an update clause, a unit subquery call, or a procedure call with no YIELD.",
      p(19, 1, 20)
    )
  }

  test("Scoped Subquery with only USE") {
    run("WITH 1 AS a CALL () { USE x } RETURN a", withMultiGraphs).hasError(
      getGql42001_42N71(22, 1, 23),
      "Query must conclude with a RETURN clause, a FINISH clause, an update clause, a unit subquery call, or a procedure call with no YIELD.",
      p(22, 1, 23)
    )
  }

  test("Subquery with only USE") {
    val query = "WITH 1 AS a CALL { USE x } RETURN a"
    run(query, withMultiGraphs).hasError(
      getGql42001_42N71(19, 1, 20),
      "Query must conclude with a RETURN clause, a FINISH clause, an update clause, a unit subquery call, or a procedure call with no YIELD.",
      p(19, 1, 20)
    )
  }

  test("Subquery with only USE and importing WITH") {
    val query = "WITH 1 AS a CALL { USE x WITH a } RETURN a"
    run(query, withMultiGraphs).hasError(
      getGql42001_42N71(19, 1, 20),
      "Query must conclude with a RETURN clause, a FINISH clause, an update clause, a unit subquery call, or a procedure call with no YIELD.",
      p(19, 1, 20)
    )
  }

  test(
    "should allow Multiple USE referencing the same graph when UseAsSingleGraphSelector feature is set - scoped subquery"
  ) {
    val query =
      """
        |USE x
        |WITH 1 AS a
        |CALL () {
        |  USE x
        |  RETURN 2 AS b
        |}
        |RETURN *
        |""".stripMargin

    run(query, withSingleGraph).hasNoErrors
  }

  test("should allow Multiple USE referencing the same graph when UseAsSingleGraphSelector feature is set") {
    val query =
      """
        |USE x
        |WITH 1 AS a
        |CALL {
        |  USE x
        |  RETURN 2 AS b
        |}
        |RETURN *
        |""".stripMargin

    run(query, withSingleGraph).hasErrorsIn {
      case Cypher25 => Seq.empty
      case Cypher5  => Seq.empty
    }
  }

  test(
    "should allow Multiple USE with qualified identifier referencing the same graph when UseAsSingleGraphSelector feature is set - scoped subquery"
  ) {
    val query =
      """
        |USE x.y.z
        |WITH 1 AS a
        |CALL () {
        |  USE x.y.z
        |  RETURN 2 AS b
        |}
        |RETURN *
        |""".stripMargin

    run(query, withSingleGraph).hasNoErrors
  }

  test(
    "should allow Multiple USE with qualified identifier referencing the same graph when UseAsSingleGraphSelector feature is set"
  ) {
    val query =
      """
        |USE x.y.z
        |WITH 1 AS a
        |CALL {
        |  USE x.y.z
        |  RETURN 2 AS b
        |}
        |RETURN *
        |""".stripMargin

    run(query, withSingleGraph).hasErrorsIn {
      case Cypher25 => Seq.empty
      case Cypher5  => Seq.empty
    }
  }

  test(
    "should not allow Multiple USE referencing different graphs when UseAsSingleGraphSelector feature is set - scoped subquery"
  ) {
    val query =
      """
        |USE x
        |WITH 1 AS a
        |CALL () {
        |  USE y
        |  RETURN 2 AS b
        |}
        |RETURN *
        |""".stripMargin

    run(query, withSingleGraph).hasError(
      gql42NA5(p(31, 5, 3)),
      messageProvider.createMultipleGraphReferencesError("y"),
      p(31, 5, 3)
    )
  }

  test("should not allow Multiple USE referencing different graphs when UseAsSingleGraphSelector feature is set") {
    val query =
      """
        |USE x
        |WITH 1 AS a
        |CALL {
        |  USE y
        |  RETURN 2 AS b
        |}
        |RETURN *
        |""".stripMargin

    run(query, withSingleGraph).hasError(
      gql42NA5(p(28, 5, 3)),
      messageProvider.createMultipleGraphReferencesError("y"),
      p(28, 5, 3)
    )
  }

  test("Allow view invocation in USE when UseAsMultipleGraphsSelector feature is set in scoped subquery") {
    val query =
      """
        |WITH 1 AS g, 2 AS k
        |CALL (g, k) {
        |  USE graph.byName(g, w(k))
        |  RETURN 1 AS a
        |}
        |RETURN a
        |""".stripMargin
    run(query, withMultiGraphs).hasNoErrors
  }

  test("Allow view invocation in USE when UseAsMultipleGraphsSelector feature is set") {
    val query =
      """
        |WITH 1 AS g, 2 AS k
        |CALL {
        |  USE graph.byName(g, w(k))
        |  RETURN 1 AS a
        |}
        |RETURN a
        |""".stripMargin
    run(query, withMultiGraphs).hasErrorsIn {
      case Cypher25 => Seq.empty
      case Cypher5  => Seq.empty
    }
  }

  test("Don't allow view invocation in USE when UseAsSingleGraphSelector feature is set in scoped subquery") {
    val query =
      """
        |WITH 1 AS g, 2 AS k
        |CALL (g, k) {
        |  USE graph.byName(g, w(k))
        |  RETURN 1 AS a
        |}
        |RETURN a
        |""".stripMargin
    run(query, withSingleGraph).hasError(
      GqlHelper.getGql42001_42N72(37, 4, 3),
      messageProvider.createDynamicGraphReferenceUnsupportedError("graph.byName(g, w(k))"),
      p(37, 4, 3)
    )
  }

  test("Don't allow view invocation in USE when UseAsSingleGraphSelector feature is set") {
    val query =
      """
        |WITH 1 AS g, 2 AS k
        |CALL {
        |  USE graph.byName(g, w(k))
        |  RETURN 1 AS a
        |}
        |RETURN a
        |""".stripMargin
    run(query, withSingleGraph).hasError(
      GqlHelper.getGql42001_42N72(30, 4, 3),
      messageProvider.createDynamicGraphReferenceUnsupportedError("graph.byName(g, w(k))"),
      p(30, 4, 3)
    )
  }

  test("Allow qualified view invocation in USE in scoped subquery") {
    val query =
      """
        |WITH 1 AS g, 2 AS k
        |CALL (g, k){
        |  USE graph.byName(g, x.g(), x.v(k))
        |  RETURN 1 AS a
        |}
        |RETURN a
        |""".stripMargin
    run(query, withMultiGraphs).hasNoErrors
  }

  test("Allow qualified view invocation in USE") {
    val query =
      """
        |WITH 1 AS g, 2 AS k
        |CALL {
        |  USE graph.byName(g, x.g(), x.v(k))
        |  RETURN 1 AS a
        |}
        |RETURN a
        |""".stripMargin
    run(query, withMultiGraphs).hasErrorsIn {
      case Cypher25 => Seq.empty
      case Cypher5  => Seq.empty
    }
  }

  test("Allow expressions in view invocations (with feature flag) in scoped subquery") {
    val query =
      """
        |WITH 1 AS x
        |CALL (x) {
        |  USE graph.byName(2, 'x', x, x+3)
        |  RETURN 1 AS a
        |}
        |RETURN a
        |""".stripMargin
    run(query, withMultiGraphs).hasNoErrors
  }

  test("Allow expressions in view invocations (with feature flag)") {
    val query =
      """
        |WITH 1 AS x
        |CALL {
        |  USE graph.byName(2, 'x', x, x+3)
        |  RETURN 1 AS a
        |}
        |RETURN a
        |""".stripMargin
    run(query, withMultiGraphs).hasErrorsIn {
      case Cypher25 => Seq.empty
      case Cypher5  => Seq.empty
    }
  }

  test("Expressions in view invocations are checked (with feature flag) in scoped subquery") {
    val query =
      """
        |WITH 1 AS x
        |CALL () {
        |  USE graph.byName(2, 'x', y, x+3)
        |  RETURN 1 AS a
        |}
        |RETURN a
        |""".stripMargin

    run(query, withMultiGraphs).hasErrors(
      SemanticError(
        ErrorGqlStatusObjectImplementation.from(GqlStatusInfoCodes.STATUS_42001)
          .atPosition(50, 4, 28)
          .withCause(ErrorGqlStatusObjectImplementation.from(GqlStatusInfoCodes.STATUS_42N62)
            .atPosition(50, 4, 28)
            .withParam(GqlParams.StringParam.variable, "y")
            .build())
          .build(),
        "Variable `y` not defined",
        p(50, 4, 28)
      ),
      SemanticError(
        ErrorGqlStatusObjectImplementation.from(GqlStatusInfoCodes.STATUS_42001)
          .atPosition(53, 4, 31)
          .withCause(ErrorGqlStatusObjectImplementation.from(GqlStatusInfoCodes.STATUS_42N62)
            .atPosition(53, 4, 31)
            .withParam(GqlParams.StringParam.variable, "x")
            .build())
          .build(),
        "Variable `x` not defined",
        p(53, 4, 31)
      )
    )
  }

  test("Expressions in view invocations are checked (with feature flag)") {
    val query =
      """
        |WITH 1 AS x
        |CALL {
        |  USE graph.byName(2, 'x', y, x+3)
        |  RETURN 1 AS a
        |}
        |RETURN a
        |""".stripMargin

    run(query, withMultiGraphs).hasError(
      ErrorGqlStatusObjectImplementation.from(GqlStatusInfoCodes.STATUS_42001)
        .atPosition(47, 4, 28)
        .withCause(ErrorGqlStatusObjectImplementation.from(GqlStatusInfoCodes.STATUS_42N62)
          .atPosition(47, 4, 28)
          .withParam(GqlParams.StringParam.variable, "y")
          .build())
        .build(),
      "Variable `y` not defined",
      p(47, 4, 28)
    )
  }

  test("should allow USE only in leading sub-query position in scoped subquery") {
    val query =
      """
        |WITH 1 AS x
        |CALL () {
        |  MATCH (n)
        |  USE g
        |  RETURN n
        |}
        |RETURN n
        |""".stripMargin
    run(query, withSingleGraph).hasError(
      SemanticError.invalidPlacementOfUseClause(
        p(37, 5, 3)
      ).gqlStatusObject,
      "USE clause must be the first clause in a (sub-)query.",
      p(37, 5, 3)
    )
  }

  test("should allow USE only in leading sub-query position") {
    val query =
      """
        |WITH 1 AS x
        |CALL {
        |  MATCH (n)
        |  USE g
        |  RETURN n
        |}
        |RETURN n
        |""".stripMargin
    run(query, withSingleGraph).hasError(
      GqlHelper.getGql42001_42N73(34, 5, 3),
      "USE clause must be either the first clause in a (sub-)query or preceded by an importing WITH clause in a sub-query.",
      p(34, 5, 3)
    )
  }

  test("should not allow non-importing WITH before USE in sub-query position") {
    val query =
      """
        |WITH 1 AS x
        |CALL {
        |  WITH 1 AS y
        |  USE g
        |  RETURN 1 AS a
        |}
        |RETURN a
        |""".stripMargin
    run(query, withSingleGraph).hasError(
      GqlHelper.getGql42001_42N73(36, 5, 3),
      "USE clause must be either the first clause in a (sub-)query or preceded by an importing WITH clause in a sub-query.",
      p(36, 5, 3)
    )
  }

  test("should allow importing WITH before USE in sub-query position") {
    val query =
      """
        |WITH 1 AS x
        |CALL {
        |  WITH x
        |  USE g
        |  RETURN 1 AS a
        |}
        |RETURN a
        |""".stripMargin
    run(query, withSingleGraph).hasErrorsIn {
      case Cypher25 => Seq.empty
      case Cypher5  => Seq.empty
    }
  }

  test("Scoped Subquery with only MATCH") {
    val query = "WITH 1 AS a CALL () { MATCH (n) } RETURN a"
    run(query).hasError(
      getGql42001_42N71(22, 1, 23),
      "Query cannot conclude with MATCH (must be a RETURN clause, a FINISH clause, an update clause, a unit subquery call, or a procedure call with no YIELD).",
      p(22, 1, 23)
    )
  }

  test("Subquery with only MATCH") {
    val query = "WITH 1 AS a CALL { MATCH (n) } RETURN a"
    run(query).hasError(
      getGql42001_42N71(19, 1, 20),
      "Query cannot conclude with MATCH (must be a RETURN clause, a FINISH clause, an update clause, a unit subquery call, or a procedure call with no YIELD).",
      p(19, 1, 20)
    )
  }

  // Utilities for the following tests
  private val returnStarNMCombinations = Seq(
    "n",
    "m",
    "n, m",
    "*",
    "*, n",
    "*, m",
    "*, n, m"
  )

  // Utilities for the following tests
  private val returnStarOrderedNMCombinations = Seq(
    "n",
    "m",
    "n, m",
    "*"
  )

  private def containedVariables(returnStarNMCombination: String): Set[String] = {
    val strings = returnStarNMCombination.split(',').map(_.trim)
    if (strings.contains("*")) {
      Set("n", "m")
    } else {
      strings.toSet
    }
  }

  test("RETURN * in a CALL () should export variables") {
    for {
      subqueryReturn <- returnStarNMCombinations
      subqueryReturnVars = containedVariables(subqueryReturn)
      finalReturn <- returnStarNMCombinations
      finalReturnVars = containedVariables(finalReturn)
      if finalReturnVars.subsetOf(subqueryReturnVars)
    } {
      val query =
        s"""
           |CALL () {
           |  MATCH (n), (m)
           |  RETURN $subqueryReturn
           |}
           |RETURN $finalReturn
           |""".stripMargin
      run(query)
        .hasNoErrors
        .assert { result =>
          val lastClause = result.state.statement().asInstanceOf[SingleQuery].clauses.last
          result.state.semantics().scope(lastClause).get.symbolNames shouldBe finalReturnVars
        }
    }
  }

  test("RETURN * in UNION in a CALL should export variables") {
    for {
      firstSubqueryReturn <- returnStarNMCombinations
      firstSubqueryReturnVars = containedVariables(firstSubqueryReturn)
      secondSubqueryReturn <- returnStarNMCombinations
      secondSubqueryReturnVars = containedVariables(secondSubqueryReturn)
      if secondSubqueryReturnVars == firstSubqueryReturnVars
      finalReturn <- returnStarNMCombinations
      finalReturnVars = containedVariables(finalReturn)
      if finalReturnVars.subsetOf(firstSubqueryReturnVars)
    } {
      val query =
        s"""
           |CALL {
           |  MATCH (n), (m)
           |  RETURN $firstSubqueryReturn
           |    UNION
           |  MATCH (n), (m)
           |  RETURN $secondSubqueryReturn
           |}
           |RETURN $finalReturn
           |""".stripMargin
      run(query)
        .assertIn {
          case Cypher25 => _ => // No assertion
          case Cypher5  => res => res.errors shouldBe empty
        }
        .assert { result =>
          val lastClause = result.state.statement().asInstanceOf[SingleQuery].clauses.last
          result.state.semantics().scope(lastClause).get.symbolNames shouldBe finalReturnVars
        }
    }
  }

  test("RETURN * in UNION in a CALL () should export variables - with matching return columns") {
    for {
      firstSubqueryReturn <- returnStarOrderedNMCombinations
      firstSubqueryReturnVars = containedVariables(firstSubqueryReturn)
      secondSubqueryReturn <- returnStarOrderedNMCombinations
      secondSubqueryReturnVars = containedVariables(secondSubqueryReturn)
      if secondSubqueryReturnVars == firstSubqueryReturnVars
      finalReturn <- returnStarOrderedNMCombinations
      finalReturnVars = containedVariables(finalReturn)
      if finalReturnVars.subsetOf(firstSubqueryReturnVars)
    } {
      val query =
        s"""
           |CALL () {
           |  MATCH (n), (m)
           |  RETURN $firstSubqueryReturn
           |    UNION
           |  MATCH (n), (m)
           |  RETURN $secondSubqueryReturn
           |}
           |RETURN $finalReturn
           |""".stripMargin
      run(query)
        .hasNoErrors
        .assert { result =>
          val lastClause = result.state.statement().asInstanceOf[SingleQuery].clauses.last
          result.state.semantics().scope(lastClause).get.symbolNames shouldBe finalReturnVars
        }
    }
  }

  test("RETURN * in 3-way-UNION in a CALL should export variables") {
    for {
      firstSubqueryReturn <- returnStarNMCombinations
      firstSubqueryReturnVars = containedVariables(firstSubqueryReturn)
      secondSubqueryReturn <- returnStarNMCombinations
      secondSubqueryReturnVars = containedVariables(secondSubqueryReturn)
      if secondSubqueryReturnVars == firstSubqueryReturnVars
      thirdSubqueryReturn <- returnStarNMCombinations
      thirdSubqueryReturnVars = containedVariables(thirdSubqueryReturn)
      if thirdSubqueryReturnVars == firstSubqueryReturnVars
      finalReturn <- returnStarNMCombinations
      finalReturnVars = containedVariables(finalReturn)
      if finalReturnVars.subsetOf(firstSubqueryReturnVars)
    } {
      val query =
        s"""
           |CALL {
           |  MATCH (n), (m)
           |  RETURN $firstSubqueryReturn
           |    UNION
           |  MATCH (n), (m)
           |  RETURN $secondSubqueryReturn
           |    UNION
           |  MATCH (n), (m)
           |  RETURN $thirdSubqueryReturn
           |}
           |RETURN $finalReturn
           |""".stripMargin

      run(query)
        .assertIn {
          case Cypher25 => _ => // No assertion
          case Cypher5  => _.errors shouldBe empty
        }
        .assert { result =>
          val lastClause = result.state.statement().asInstanceOf[SingleQuery].clauses.last
          result.state.semantics().scope(lastClause).get.symbolNames shouldBe finalReturnVars
        }
    }
  }

  test("RETURN * in 3-way-UNION in a CALL () should export variables - with matching return columns") {
    for {
      firstSubqueryReturn <- returnStarOrderedNMCombinations
      firstSubqueryReturnVars = containedVariables(firstSubqueryReturn)
      secondSubqueryReturn <- returnStarOrderedNMCombinations
      secondSubqueryReturnVars = containedVariables(secondSubqueryReturn)
      if secondSubqueryReturnVars == firstSubqueryReturnVars
      thirdSubqueryReturn <- returnStarOrderedNMCombinations
      thirdSubqueryReturnVars = containedVariables(thirdSubqueryReturn)
      if thirdSubqueryReturnVars == firstSubqueryReturnVars
      finalReturn <- returnStarOrderedNMCombinations
      finalReturnVars = containedVariables(finalReturn)
      if finalReturnVars.subsetOf(firstSubqueryReturnVars)
    } {
      val query =
        s"""
           |CALL () {
           |  MATCH (n), (m)
           |  RETURN $firstSubqueryReturn
           |    UNION
           |  MATCH (n), (m)
           |  RETURN $firstSubqueryReturn
           |    UNION
           |  MATCH (n), (m)
           |  RETURN $thirdSubqueryReturn
           |}
           |RETURN $finalReturn
           |""".stripMargin

      run(query)
        .hasNoErrors
        .assert { result =>
          val lastClause = result.state.statement().asInstanceOf[SingleQuery].clauses.last
          result.state.semantics().scope(lastClause).get.symbolNames shouldBe finalReturnVars
        }
    }
  }

  override def messageProvider: ErrorMessageProvider = new ErrorMessageProviderAdapter {

    override def createDynamicGraphReferenceUnsupportedError(graphName: String): String =
      "A very nice message explaining why dynamic graph references are not allowed: " + graphName

    override def createMultipleGraphReferencesError(graphName: String, transactionalDefault: Boolean = false): String =
      "A very nice message explaining why multiple graph references are not allowed: " + graphName
  }
}
