/*
 * Copyright (c) "Neo4j"
 * Neo4j Sweden AB [https://neo4j.com]
 *
 * This file is part of Neo4j.
 *
 * Neo4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package org.neo4j.cypher.internal.runtime.interpreted.commands.expressions

import org.neo4j.cypher.internal.runtime.ReadableRow
import org.neo4j.cypher.internal.runtime.interpreted.pipes.QueryState
import org.neo4j.cypher.internal.runtime.interpreted.pipes.aggregation.AggregationFunction
import org.neo4j.cypher.internal.util.symbols.CypherType
import org.neo4j.exceptions.SyntaxException
import org.neo4j.memory.MemoryTracker
import org.neo4j.values.AnyValue

abstract class AggregationExpression extends Expression {

  def apply(row: ReadableRow, state: QueryState): AnyValue =
    throw new UnsupportedOperationException("Aggregations should not be used like this.")

  def createAggregationFunction(memoryTracker: MemoryTracker): AggregationFunction
}

abstract class AggregationWithInnerExpression(inner: Expression) extends AggregationExpression {

  if (inner.containsAggregate)
    // There is already a check for this in semantic checking so we will never come here
    throw SyntaxException.invalidUseOfAggregateFunction(
      "aggregate",
      "Can't use aggregate functions inside of aggregate functions."
    )

  if (!inner.isDeterministic)
    throw SyntaxException.invalidUseOfAggregateFunction(
      "non-deterministic",
      "Can't use non-deterministic (random) functions inside of aggregate functions."
    )

  def expectedInnerType: CypherType

  def arguments: Seq[Expression] = Seq(inner)

}
