/*
 * Copyright (c) "Neo4j"
 * Neo4j Sweden AB [https://neo4j.com]
 *
 * This file is part of Neo4j.
 *
 * Neo4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package org.neo4j.internal.batchimport.input.parquet;

import java.io.Closeable;
import java.io.IOException;
import java.time.ZoneId;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Supplier;
import org.apache.parquet.ParquetReadOptions;
import org.apache.parquet.column.ColumnDescriptor;
import org.apache.parquet.column.ColumnReader;
import org.apache.parquet.column.impl.ColumnReadStoreImpl;
import org.apache.parquet.column.page.PageReadStore;
import org.apache.parquet.example.DummyRecordConverter;
import org.apache.parquet.hadoop.ParquetFileReader;
import org.apache.parquet.io.api.GroupConverter;
import org.apache.parquet.schema.GroupType;
import org.apache.parquet.schema.LogicalTypeAnnotation;
import org.apache.parquet.schema.MessageType;
import org.apache.parquet.schema.PrimitiveType;
import org.apache.parquet.schema.Type;
import org.neo4j.batchimport.api.input.IdType;
import org.neo4j.internal.batchimport.input.Groups;

/**
 * There should be a 1:1 match between Reader and file for now.
 */
class ParquetDataReader implements Closeable {

    private final ParquetData parquetDataFile;
    private final Groups groups;
    private final IdType idType;
    private final Supplier<ZoneId> defaultTimezoneSupplier;
    private final String arrayDelimiter;
    private final ParquetFileReader reader;
    private final AtomicInteger blockCounter;
    private final MessageType schema;
    private final GroupConverter recordConverter;
    private final String createdBy;
    private final List<ColumnDescriptor> parquetColumns;

    ParquetDataReader(
            ParquetData parquetDataFile,
            Groups groups,
            IdType idType,
            Supplier<ZoneId> defaultTimezoneSupplier,
            String arrayDelimiter) {
        this.parquetDataFile = parquetDataFile;
        this.groups = groups;
        this.idType = idType;
        this.defaultTimezoneSupplier = defaultTimezoneSupplier;
        this.arrayDelimiter = arrayDelimiter;
        var path = parquetDataFile.file();

        try {
            this.reader = ParquetFileReader.open(
                    ParquetInput.ParquetImportInputFile.of(path),
                    ParquetReadOptions.builder().build());
            var metadata = this.reader.getFileMetaData();
            this.schema = metadata.getSchema();
            this.recordConverter = new DummyRecordConverter(this.schema).getRootConverter();
            this.createdBy = metadata.getCreatedBy();

            var columnsToRead = parquetDataFile.columns().stream()
                    .map(column -> column.headerDefinition().parquetColumnName())
                    .toList();
            this.parquetColumns = schema.getColumns().stream()
                    .filter(c -> columnsToRead.contains(c.getPath()[0]))
                    .toList();

            this.blockCounter = new AtomicInteger(0);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    public Iterator<List<Object>> next() throws IOException {
        var nextRowGroupIndex = blockCounter.getAndIncrement();
        if (nextRowGroupIndex >= reader.getRowGroups().size()) {
            return null;
        }

        return new ParquetRowGroupReader(reader.readRowGroup(nextRowGroupIndex));
    }

    public ParquetData getParquetDataFile() {
        return parquetDataFile;
    }

    public Groups getGroups() {
        return groups;
    }

    public IdType getIdType() {
        return idType;
    }

    public Supplier<ZoneId> getDefaultTimezoneSupplier() {
        return defaultTimezoneSupplier;
    }

    public String getArrayDelimiter() {
        return arrayDelimiter;
    }

    private class ParquetRowGroupReader implements Iterator<List<Object>> {
        private final List<ColumnReader> columnReaders;
        private final long rowCount;
        private long rowIndex;

        ParquetRowGroupReader(PageReadStore store) {
            this.rowCount = store.getRowCount();

            var columnReadStore = new ColumnReadStoreImpl(
                    store,
                    ParquetDataReader.this.recordConverter,
                    ParquetDataReader.this.schema,
                    ParquetDataReader.this.createdBy);
            this.columnReaders = parquetColumns.stream()
                    .map(columnReadStore::getColumnReader)
                    .toList();
        }

        @Override
        public boolean hasNext() {
            return rowIndex < rowCount;
        }

        // Wrapper class for representing maps / structs and their
        // content.
        private static class MapLikeRecord {

            private final String fieldName;
            private Keys keys;
            private Values values;

            private MapLikeRecord(String fieldName) {
                this.fieldName = fieldName;
            }

            private MapLikeRecord(String fieldName, Keys keys, Values values) {
                this.fieldName = fieldName;
                this.keys = keys;
                this.values = values;
            }

            // Avoid passing around a custom type to the ParquetDataInputChunk
            // by merging it down to a simple map.
            private Map<String, Object> asMap() {
                Map<String, Object> renderedMap = new HashMap<>();
                for (int i = 0; i < keys.keys().size(); i++) {
                    renderedMap.put(keys.keys.get(i), values.values().get(i));
                }

                return renderedMap;
            }

            // marker classes for map keys to be picked up by the next function
            private record Keys(String fieldName, List<String> keys) {}

            // marker classes for map values to be picked up by the next function
            private record Values(String fieldName, List<Object> values) {}
        }

        @Override
        public List<Object> next() {
            var result = new ArrayList<>();
            var processedEmptyMaps = new HashSet<String>();
            MapLikeRecord mapLikeRecord = null;
            for (ColumnReader columnReader : this.columnReaders) {
                Object readValue = readValue(columnReader);
                // we only get a direct Map value if it's empty
                if (readValue instanceof Map emptyRecord) {
                    if (!processedEmptyMaps.contains(
                            columnReader.getDescriptor().getPath()[0])) {
                        result.add(emptyRecord);
                        processedEmptyMaps.add(columnReader.getDescriptor().getPath()[0]);
                    }
                } else if (readValue instanceof MapLikeRecord.Keys keys) {
                    if (mapLikeRecord == null) {
                        mapLikeRecord =
                                new MapLikeRecord(columnReader.getDescriptor().getPath()[0]);
                    } else if (!mapLikeRecord.fieldName.equals(keys.fieldName)) {
                        result.add(mapLikeRecord.asMap());
                        mapLikeRecord = new MapLikeRecord(keys.fieldName);
                    }
                    mapLikeRecord.keys = keys;
                } else if (readValue instanceof MapLikeRecord.Values values) {
                    if (mapLikeRecord == null) {
                        mapLikeRecord =
                                new MapLikeRecord(columnReader.getDescriptor().getPath()[0]);
                    } else if (!mapLikeRecord.fieldName.equals(values.fieldName)) {
                        result.add(mapLikeRecord.asMap());
                        mapLikeRecord = new MapLikeRecord(values.fieldName);
                    }
                    mapLikeRecord.values = values;
                } else if (readValue instanceof MapLikeRecord struct) {
                    if (mapLikeRecord != null && mapLikeRecord.keys != null) {
                        result.add(mapLikeRecord.asMap());
                        mapLikeRecord = null;
                    }
                    result.add(struct.asMap());

                } else {
                    // when switching to a non-map-related column, we can be sure that
                    // if there was a map before, it was read completely.
                    // add mapRecord if there is something upfront setting the next value
                    if (mapLikeRecord != null && mapLikeRecord.keys != null) {
                        result.add(mapLikeRecord.asMap());
                        mapLikeRecord = null;
                    }
                    result.add(readValue);
                }
            }
            // add if map record is last
            if (mapLikeRecord != null && mapLikeRecord.keys != null) {
                result.add(mapLikeRecord.asMap());
            }

            this.rowIndex++;

            return result;
        }

        /* This method does the `consume` call on the columnReader because the list
         * processing requires it for every single value.
         */
        private Object readValue(ColumnReader columnReader) {
            ColumnDescriptor column = columnReader.getDescriptor();
            PrimitiveType primitiveType = column.getPrimitiveType();
            int maxDefinitionLevel = column.getMaxDefinitionLevel();
            String fieldName = column.getPath()[0];
            Type type = schema.getType(fieldName);
            LogicalTypeAnnotation logicalType = type.getLogicalTypeAnnotation();
            if (columnReader.getCurrentDefinitionLevel() == maxDefinitionLevel) {
                if (logicalType != null && logicalType.equals(LogicalTypeAnnotation.listType())) {
                    var readValues = new ArrayList<>();
                    for (var i = 0; i < columnReader.getTotalValueCount(); i++) {
                        readValues.add(readPrimitiveType(columnReader, primitiveType));
                        columnReader.consume();
                    }
                    return readValues;
                }
                if (logicalType != null && logicalType.equals(LogicalTypeAnnotation.mapType())) {
                    // Initialize data structure for a map by fieldName
                    // the data will come in as "all keys" and then "all values".
                    // As a consequence, we need to return them separately and merge them
                    // later into a Java map.
                    if (column.getPath()[2].equals("key")) {
                        var mapKeys = new MapLikeRecord.Keys(fieldName, new ArrayList<>());
                        for (var i = 0; i < columnReader.getTotalValueCount(); i++) {
                            String key = (String) readPrimitiveType(columnReader, primitiveType);
                            mapKeys.keys().add(fieldName + "." + key);
                            columnReader.consume();
                        }
                        return mapKeys;
                    } else {
                        var mapValues = new MapLikeRecord.Values(fieldName, new ArrayList<>());
                        for (var i = 0; i < columnReader.getTotalValueCount(); i++) {
                            Object value = readPrimitiveType(columnReader, primitiveType);
                            mapValues.values().add(value);
                            columnReader.consume();
                        }
                        return mapValues;
                    }
                }
                if (type instanceof GroupType) {
                    // The only groupType supported right now is a struct.
                    // This could also represent _any_ complex type.
                    var mapKeys = new MapLikeRecord.Keys(fieldName, new ArrayList<>());
                    String propertyName = fieldName + "." + column.getPath()[1];
                    mapKeys.keys().add(propertyName);
                    var mapValues = new MapLikeRecord.Values(fieldName, new ArrayList<>());
                    mapValues.values().add(readPrimitiveType(columnReader, primitiveType));
                    return new MapLikeRecord(fieldName, mapKeys, mapValues);
                } else {
                    Object readValue = readPrimitiveType(columnReader, primitiveType);
                    columnReader.consume();
                    return readValue;
                }
            } else {
                columnReader.consume();
                if (logicalType != null) {
                    if (logicalType.equals(LogicalTypeAnnotation.listType())) {
                        return List.of();
                    }
                    if (logicalType.equals(LogicalTypeAnnotation.mapType())) {
                        return Map.of();
                    }
                }
                return null;
            }
        }

        private static Object readPrimitiveType(ColumnReader columnReader, PrimitiveType primitiveType) {
            return switch (primitiveType.getPrimitiveTypeName()) {
                case BINARY, FIXED_LEN_BYTE_ARRAY, INT96 -> primitiveType
                        .stringifier()
                        .stringify(columnReader.getBinary());
                case BOOLEAN -> columnReader.getBoolean();
                case DOUBLE -> columnReader.getDouble();
                case FLOAT -> columnReader.getFloat();
                case INT32 -> columnReader.getInteger();
                case INT64 -> columnReader.getLong();
            };
        }
    }

    @Override
    public void close() throws IOException {
        this.reader.close();
    }
}
