/*
 * Copyright (c) "Neo4j"
 * Neo4j Sweden AB [https://neo4j.com]
 *
 * This file is part of Neo4j.
 *
 * Neo4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package org.neo4j.dbms.systemgraph;

import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;
import org.neo4j.cypher.internal.CypherVersion;
import org.neo4j.graphdb.Direction;
import org.neo4j.graphdb.Node;
import org.neo4j.graphdb.ResourceIterator;
import org.neo4j.graphdb.Transaction;
import org.neo4j.internal.helpers.collection.Pair;
import org.neo4j.kernel.database.DatabaseReference;
import org.neo4j.kernel.database.DatabaseReferenceImpl;
import org.neo4j.kernel.database.NamedDatabaseId;
import org.neo4j.kernel.database.NormalizedCatalogEntry;
import org.neo4j.kernel.database.NormalizedDatabaseName;

public class CommunityTopologyGraphDbmsModel implements TopologyGraphDbmsModel {
    protected final Transaction tx;

    public CommunityTopologyGraphDbmsModel(Transaction tx) {
        this.tx = tx;
    }

    public Map<NamedDatabaseId, TopologyGraphDbmsModel.DatabaseAccess> getAllDatabaseAccess() {
        try (ResourceIterator<Node> nodes = tx.findNodes(DATABASE_LABEL)) {
            return nodes.stream()
                    .collect(Collectors.toMap(
                            CommunityTopologyGraphDbmsModelUtil::getDatabaseId,
                            CommunityTopologyGraphDbmsModelUtil::getDatabaseAccess));
        }
    }

    @Override
    public Optional<NamedDatabaseId> getDatabaseIdByAlias(String databaseName) {
        return CommunityTopologyGraphDbmsModelUtil.getDatabaseIdByAliasInRoot(tx, databaseName)
                .or(() ->
                        CommunityTopologyGraphDbmsModelUtil.getDatabaseIdBy(tx, DATABASE_NAME_PROPERTY, databaseName));
    }

    @Override
    public Optional<NamedDatabaseId> getDatabaseIdByUUID(UUID uuid) {
        return CommunityTopologyGraphDbmsModelUtil.getDatabaseIdBy(tx, DATABASE_UUID_PROPERTY, uuid.toString());
    }

    @Override
    public Set<DatabaseReference> getAllDatabaseReferences() {
        var primaryRefs = CommunityTopologyGraphDbmsModelUtil.getAllPrimaryStandardDatabaseReferencesInRoot(tx);
        var internalAliasRefs = getAllInternalDatabaseReferencesInRoot();
        var externalRefs = getAllExternalDatabaseReferencesInRoot();
        var compositeRefs = getAllCompositeDatabaseReferencesInRoot();
        var spdGraphShardRefs = getAllSPDGraphShardReferencesInRoot();
        var spdPropertyShardRefs = getAllSPDPropertyShardReferencesInRoot();
        var mirrorRefs = getAllMirrorReferencesInRoot();
        return Stream.of(
                        primaryRefs,
                        internalAliasRefs,
                        externalRefs,
                        compositeRefs,
                        spdGraphShardRefs,
                        spdPropertyShardRefs,
                        mirrorRefs)
                .flatMap(s -> s)
                .collect(Collectors.toUnmodifiableSet());
    }

    @Override
    public Set<DatabaseReferenceImpl.Composite> getAllCompositeDatabaseReferences() {
        return getAllCompositeDatabaseReferencesInRoot().collect(Collectors.toUnmodifiableSet());
    }

    @Override
    public Optional<DatabaseReference> getDatabaseRefByAlias(NormalizedCatalogEntry catalogEntry) {
        if (catalogEntry.compositeDb().isPresent()) {
            return resolveConstituent(catalogEntry.compositeDb().get(), catalogEntry.databaseAlias());
        } else {
            return resolveRootReference(catalogEntry.databaseAlias());
        }
    }

    private Optional<DatabaseReference> resolveConstituent(String composite, String constituent) {
        return Optional.<DatabaseReference>empty()
                .or(() -> CommunityTopologyGraphDbmsModelUtil.getInternalDatabaseReferenceInRoot(
                        tx, composite, constituent))
                .or(() -> CommunityTopologyGraphDbmsModelUtil.getExternalDatabaseReferenceInRoot(
                        tx, composite, constituent));
    }

    private Optional<DatabaseReference> resolveRootReference(String normalizedDatabaseAlias) {
        return Optional.<DatabaseReference>empty()
                .or(() -> getCompositeDatabaseReferenceInRoot(normalizedDatabaseAlias))
                .or(() -> getSPDGraphShardReferenceInRoot(normalizedDatabaseAlias))
                .or(() -> getSPDPropertyShardReferenceInRoot(normalizedDatabaseAlias))
                .or(() -> getMirrorReferencesInRoot(normalizedDatabaseAlias))
                .or(() -> CommunityTopologyGraphDbmsModelUtil.getInternalDatabaseReferenceInRoot(
                        tx, normalizedDatabaseAlias))
                .or(() -> CommunityTopologyGraphDbmsModelUtil.getExternalDatabaseReferenceInRoot(
                        tx, normalizedDatabaseAlias));
    }

    private Stream<Node> getAllAliasNodesInRoot() {
        return getAllAliasNodesInNamespace(DEFAULT_NAMESPACE);
    }

    private Stream<Node> getAllAliasNodesInNamespace(String namespace) {
        return tx.findNodes(DATABASE_NAME_LABEL, NAMESPACE_PROPERTY, namespace).stream();
    }

    private Stream<Node> getRemoteAliasNodesInNamespace(String namespace) {
        return tx.findNodes(REMOTE_DATABASE_LABEL, NAMESPACE_PROPERTY, namespace).stream();
    }

    private Stream<Node> getAliasNodeInNamespace(String namespace, String databaseName) {
        return tx.findNodes(DATABASE_NAME_LABEL, NAMESPACE_PROPERTY, namespace, NAME_PROPERTY, databaseName).stream();
    }

    private Stream<Node> getAliasNodeInRoot(String databaseName) {
        return getAliasNodeInNamespace(DEFAULT_NAMESPACE, databaseName);
    }

    @Override
    public Optional<DriverSettings> getDriverSettings(String databaseName, String namespace) {
        databaseName = NormalizedDatabaseName.normalize(databaseName);
        namespace = NormalizedDatabaseName.normalize(namespace);
        return tx.findNodes(REMOTE_DATABASE_LABEL, NAME_PROPERTY, databaseName, NAMESPACE_PROPERTY, namespace).stream()
                .findFirst()
                .flatMap(CommunityTopologyGraphDbmsModelUtil::getDriverSettings);
    }

    @Override
    public Optional<Map<String, Object>> getAliasProperties(String databaseName, String namespace) {
        databaseName = NormalizedDatabaseName.normalize(databaseName);
        namespace = NormalizedDatabaseName.normalize(namespace);
        return getAliasNodeInNamespace(namespace, databaseName)
                .findFirst()
                .flatMap(CommunityTopologyGraphDbmsModelUtil::getAliasProperties);
    }

    @Override
    public Optional<CypherVersion> getRemoteAliasLanguageVersion(String remoteAliasName) {
        try (var nodes = tx.findNodes(REMOTE_DATABASE_LABEL, NAME_PROPERTY, remoteAliasName)) {
            var filtered = nodes.stream()
                    .filter(node -> node.getProperty(NAMESPACE_PROPERTY).equals(DEFAULT_NAMESPACE))
                    .toList();
            if (filtered.isEmpty()) {
                return Optional.empty();
            }
            String defaultLanguage = (String) filtered.getFirst().getProperty(DATABASE_DEFAULT_LANGUAGE_PROPERTY);
            if (defaultLanguage != null) {
                return CypherVersion.fromStoredValueOptional(defaultLanguage);
            } else {
                return Optional.empty();
            }
        }
    }

    @Override
    public Optional<ExternalDatabaseCredentials> getExternalDatabaseCredentials(
            DatabaseReferenceImpl.External databaseReference) {
        String databaseName = databaseReference.alias().name();
        String namespace =
                databaseReference.namespace().map(NormalizedDatabaseName::name).orElse(DEFAULT_NAMESPACE);
        return tx.findNodes(REMOTE_DATABASE_LABEL, NAME_PROPERTY, databaseName, NAMESPACE_PROPERTY, namespace).stream()
                .findFirst()
                .flatMap(CommunityTopologyGraphDbmsModelUtil::getDatabaseCredentials);
    }

    private Stream<DatabaseReferenceImpl.Mirror> getAllMirrorReferencesInRoot() {
        return getAllAliasNodesInRoot()
                .flatMap(alias -> CommunityTopologyGraphDbmsModelUtil.getTargetedDatabaseNode(alias)
                        .filter(db -> db.getDegree(IS_MIRROR_OF_RELATIONSHIP, Direction.OUTGOING) > 0)
                        .flatMap(db -> createMirrorReference(alias, db))
                        .stream());
    }

    private Optional<DatabaseReferenceImpl.Mirror> getMirrorReferencesInRoot(String databaseName) {
        return getAliasNodeInRoot(databaseName)
                .flatMap(alias -> CommunityTopologyGraphDbmsModelUtil.getTargetedDatabaseNode(alias)
                        .filter(db -> db.getDegree(IS_MIRROR_OF_RELATIONSHIP, Direction.OUTGOING) > 0)
                        .flatMap(db -> createMirrorReference(alias, db))
                        .stream())
                .findFirst();
    }

    private Stream<DatabaseReferenceImpl.Composite> getAllCompositeDatabaseReferencesInRoot() {
        return getAllAliasNodesInRoot()
                .flatMap(alias -> CommunityTopologyGraphDbmsModelUtil.getTargetedDatabaseNode(alias)
                        .filter(db -> db.hasLabel(COMPOSITE_DATABASE_LABEL))
                        .flatMap(db -> createCompositeReference(alias, db))
                        .stream());
    }

    private Optional<DatabaseReferenceImpl.Composite> getCompositeDatabaseReferenceInRoot(String databaseName) {
        return getAliasNodeInRoot(databaseName)
                .flatMap(alias -> CommunityTopologyGraphDbmsModelUtil.getTargetedDatabaseNode(alias)
                        .filter(db -> db.hasLabel(COMPOSITE_DATABASE_LABEL))
                        .flatMap(db -> createCompositeReference(alias, db))
                        .stream())
                .findFirst();
    }

    private Optional<DatabaseReferenceImpl.Composite> createCompositeReference(Node alias, Node db) {
        return CommunityTopologyGraphDbmsModelUtil.ignoreConcurrentDeletes(() -> {
            var aliasName = CommunityTopologyGraphDbmsModelUtil.getNameProperty(DATABASE_NAME, alias);
            var databaseId = CommunityTopologyGraphDbmsModelUtil.getDatabaseId(db);
            var compositeName = CommunityTopologyGraphDbmsModelUtil.getNameProperty(DATABASE, db);
            var components = getAllDatabaseReferencesInComposite(compositeName);
            return Optional.of(new DatabaseReferenceImpl.Composite(aliasName, databaseId, components));
        });
    }

    private Stream<DatabaseReferenceImpl.SPD> getAllSPDGraphShardReferencesInRoot() {
        return getAllAliasNodesInRoot()
                .flatMap(alias -> CommunityTopologyGraphDbmsModelUtil.getTargetedDatabaseNode(alias)
                        .filter(db -> db.getDegree(HAS_SHARD, Direction.OUTGOING) > 0)
                        .flatMap(db -> createSPDGraphShardReference(alias, db))
                        .stream());
    }

    private Optional<DatabaseReferenceImpl.SPD> getSPDGraphShardReferenceInRoot(String databaseName) {
        return getAliasNodeInRoot(databaseName)
                .flatMap(alias -> CommunityTopologyGraphDbmsModelUtil.getTargetedDatabaseNode(alias)
                        .filter(db -> db.getDegree(HAS_SHARD, Direction.OUTGOING) > 0)
                        .flatMap(db -> createSPDGraphShardReference(alias, db))
                        .stream())
                .findFirst();
    }

    private Stream<DatabaseReferenceImpl.SPDShard> getAllSPDPropertyShardReferencesInRoot() {
        return getAllAliasNodesInRoot()
                .flatMap(alias -> CommunityTopologyGraphDbmsModelUtil.getTargetedDatabaseNode(alias)
                        .filter(db -> db.getDegree(HAS_SHARD, Direction.INCOMING) > 0)
                        .flatMap(db -> createSPDPropertyShardReference(alias, db))
                        .stream());
    }

    private Optional<DatabaseReferenceImpl.SPDShard> getSPDPropertyShardReferenceInRoot(String databaseName) {
        return getAliasNodeInRoot(databaseName)
                .flatMap(alias -> CommunityTopologyGraphDbmsModelUtil.getTargetedDatabaseNode(alias)
                        .filter(db -> db.getDegree(HAS_SHARD, Direction.INCOMING) > 0)
                        .flatMap(db -> createSPDPropertyShardReference(alias, db))
                        .stream())
                .findFirst();
    }

    private Optional<DatabaseReferenceImpl.SPD> createSPDGraphShardReference(Node alias, Node db) {
        return CommunityTopologyGraphDbmsModelUtil.ignoreConcurrentDeletes(() -> {
            var aliasName = CommunityTopologyGraphDbmsModelUtil.getNameProperty(DATABASE_NAME, alias);
            var databaseId = CommunityTopologyGraphDbmsModelUtil.getDatabaseId(db);
            var shards = StreamSupport.stream(
                            db.getRelationships(Direction.OUTGOING, TopologyGraphDbmsModel.HAS_SHARD)
                                    .spliterator(),
                            false)
                    .flatMap(rel -> getDatabaseRefByAlias(
                            new NormalizedCatalogEntry((String) rel.getEndNode().getProperty(DATABASE_NAME_PROPERTY)))
                            .map(ref -> Pair.of((int) rel.getProperty(HAS_SHARD_INDEX_PROPERTY), ref))
                            .stream())
                    .collect(Collectors.toMap(Pair::first, Pair::other));
            return Optional.of(new DatabaseReferenceImpl.SPD(aliasName, databaseId, shards));
        });
    }

    private static Optional<DatabaseReferenceImpl.SPDShard> createSPDPropertyShardReference(Node alias, Node db) {
        return CommunityTopologyGraphDbmsModelUtil.createInternalReference(
                        alias, CommunityTopologyGraphDbmsModelUtil.getDatabaseId(db))
                .flatMap(internal -> CommunityTopologyGraphDbmsModelUtil.readOwningDatabase(db)
                        .map(internal::asShard));
    }

    private static Optional<DatabaseReferenceImpl.Mirror> createMirrorReference(Node alias, Node db) {
        return CommunityTopologyGraphDbmsModelUtil.createInternalReference(
                        alias, CommunityTopologyGraphDbmsModelUtil.getDatabaseId(db))
                .flatMap(internal -> CommunityTopologyGraphDbmsModelUtil.readUpstreamDatabase(db)
                        .map(internal::asMirror));
    }

    private Set<DatabaseReference> getAllDatabaseReferencesInComposite(NormalizedDatabaseName compositeName) {
        var internalRefs = getAllInternalDatabaseReferencesInNamespace(compositeName.name());
        var spdInternalRefs = getAllSpdDatabaseReferencesInNamespace(compositeName.name());
        var externalRefs = getAllExternalDatabaseReferencesInNamespace(compositeName.name());
        return Stream.concat(Stream.concat(internalRefs, externalRefs), spdInternalRefs)
                .collect(Collectors.toUnmodifiableSet());
    }

    private Stream<DatabaseReferenceImpl.External> getAllExternalDatabaseReferencesInRoot() {
        return getAllExternalDatabaseReferencesInNamespace(DEFAULT_NAMESPACE);
    }

    private Stream<DatabaseReferenceImpl.External> getAllExternalDatabaseReferencesInNamespace(String namespace) {
        return getRemoteAliasNodesInNamespace(namespace)
                .flatMap(alias -> CommunityTopologyGraphDbmsModelUtil.createExternalReference(alias).stream());
    }

    private Stream<DatabaseReferenceImpl.Internal> getAllInternalDatabaseReferencesInRoot() {
        return getAllInternalDatabaseReferencesInNamespace(DEFAULT_NAMESPACE);
    }

    private Stream<DatabaseReferenceImpl.Internal> getAllInternalDatabaseReferencesInNamespace(String namespace) {
        return getAllAliasNodesInNamespace(namespace)
                .flatMap(alias -> CommunityTopologyGraphDbmsModelUtil.getTargetedDatabaseNode(alias)
                        .filter(node -> !node.hasProperty(DATABASE_VIRTUAL_PROPERTY))
                        .filter(node -> node.getDegree(HAS_SHARD, Direction.OUTGOING) == 0
                                && node.getDegree(HAS_SHARD, Direction.INCOMING) == 0)
                        .filter(node -> node.getDegree(IS_MIRROR_OF_RELATIONSHIP, Direction.OUTGOING) == 0)
                        .map(CommunityTopologyGraphDbmsModelUtil::getDatabaseId)
                        .flatMap(db -> CommunityTopologyGraphDbmsModelUtil.createInternalReference(alias, db))
                        .stream());
    }

    private Stream<DatabaseReferenceImpl.Internal> getAllSpdDatabaseReferencesInNamespace(String namespace) {
        return getAllAliasNodesInNamespace(namespace)
                .flatMap(alias -> CommunityTopologyGraphDbmsModelUtil.getTargetedDatabaseNode(alias)
                        .filter(node -> !node.hasProperty(DATABASE_VIRTUAL_PROPERTY))
                        .filter(node -> node.getDegree(HAS_SHARD, Direction.OUTGOING) > 0)
                        .map(CommunityTopologyGraphDbmsModelUtil::getDatabaseId)
                        .flatMap(db -> CommunityTopologyGraphDbmsModelUtil.createInternalSpdReference(alias, db))
                        .stream());
    }
}
